"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

time.py
~~~~~~~

Functionality to add and process time varying parameters

"""
import xarray as xr
import numpy as np
import pandas as pd

from calliope import exceptions
from calliope.core.attrdict import AttrDict
from calliope.core.util.tools import plugin_load
from calliope.preprocess import checks
from calliope.core.util.dataset import reorganise_xarray_dimensions


def apply_time_clustering(model_data, model_run):
    """
    Take a Calliope model_data post time dimension addition, prior to any time
    clustering, and apply relevant time clustering/masking techniques.
    See doi: 10.1016/j.apenergy.2017.03.051 for applications.

    Techniques include:
    - Clustering timeseries into a selected number of 'representative' days.
        Days with similar profiles and daily magnitude are grouped together and
        represented by one 'representative' day with a greater weight per time
        step.
    - Masking timeseries, leading to variable timestep length
        Only certain parts of the input are shown at full resolution, with other
        periods being clustered together into a single timestep.
        E.g. Keep high resolution in the week with greatest wind power variability,
        smooth all other timesteps to 12H
    - Timestep resampling
        Used to reduce problem size by reducing resolution of all timeseries data.
        E.g. resample from 1H to 6H timesteps


    Parameters
    ----------
    model_data : xarray Dataset
        Preprocessed Calliope model_data, as produced using
        `calliope.preprocess.build_model_data`
        and found in model._model_data_original
    model_run : bool
        preprocessed model_run dictionary, as produced by
        Calliope.preprocess_model

    Returns
    -------
    data : xarray Dataset
        Dataset with optimisation parameters as variables, optimisation sets as
        coordinates, and other information in attributes. Time dimension has
        been updated as per user-defined clustering techniques (from model_run)

    """
    time_config = model_run.model["time"]

    data = model_data.copy(deep=True)

    ##
    # Process masking and get list of timesteps to keep at high res
    ##
    if "masks" in time_config:
        masks = {}
        # time.masks is a list of {'function': .., 'options': ..} dicts
        for entry in time_config.masks:
            entry = AttrDict(entry)
            mask_func = plugin_load(
                entry.function, builtin_module="calliope.time.masks"
            )
            mask_kwargs = entry.get_key("options", default=AttrDict()).as_dict()
            masks[entry.to_yaml()] = mask_func(data, **mask_kwargs)
        data.attrs["masks"] = masks
        # Concatenate the DatetimeIndexes by using dummy Series
        chosen_timesteps = pd.concat(
            [pd.Series(0, index=m) for m in masks.values()]
        ).index
        # timesteps: a list of timesteps NOT picked by masks
        timesteps = pd.Index(data.timesteps.values).difference(chosen_timesteps)
    else:
        timesteps = None

    ##
    # Process function, apply resolution adjustments
    ##
    if "function" in time_config:
        func = plugin_load(time_config.function, builtin_module="calliope.time.funcs")
        func_kwargs = time_config.get("function_options", AttrDict()).as_dict()
        if "file=" in func_kwargs.get("clustering_func", ""):
            func_kwargs.update({"model_run": model_run})
        data = func(data=data, timesteps=timesteps, **func_kwargs)

    return data


def add_time_dimension(data, model_run):
    """
    Once all constraints and costs have been loaded into the model dataset, any
    timeseries data is loaded from file and substituted into the model dataset

    Parameters:
    -----------
    data : xarray Dataset
        A data structure which has already gone through `constraints_to_dataset`,
        `costs_to_dataset`, and `add_attributes`
    model_run : AttrDict
        Calliope model_run dictionary

    Returns:
    --------
    data : xarray Dataset
        A data structure with an additional time dimension to the input dataset,
        with all relevant `file=` and `df= `entries replaced with the correct data.

    """
    data["timesteps"] = pd.to_datetime(data.timesteps.data)

    # Search through every constraint/cost for use of '='
    for variable in data.data_vars:
        # 1) If '=' in variable, it will give the variable a string data type
        if data[variable].dtype.kind != "U":
            continue

        # 2) convert to a Pandas Series to do 'string contains' search
        data_series = data[variable].to_series()

        # 3) get Series of all uses of 'file=' or 'df=' for this variable (timeseries keys)
        tskeys = data_series[
            data_series.str.contains("file=") | data_series.str.contains("df=")
        ]

        # 4) If no use of 'file=' or 'df=' then we can be on our way
        if tskeys.empty:
            continue

        # 5) remove all before '=' and split filename and location column
        tskeys = tskeys.str.split("=").str[1].str.rsplit(":", 1)
        if isinstance(tskeys.index, pd.MultiIndex):
            tskeys.index = tskeys.index.remove_unused_levels()

        # 6) Get all timeseries data from dataframes stored in model_run
        timeseries_data = []
        key_errors = []
        for loc_tech, (tskey, column) in tskeys.items():
            try:
                timeseries_data.append(
                    model_run.timeseries_data[tskey].loc[:, column].values
                )
            except KeyError:
                key_errors.append(
                    "column `{}` not found in dataframe `{}`, but was requested by "
                    "loc::tech `{}`.".format(column, tskey, loc_tech)
                )
        if key_errors:
            exceptions.print_warnings_and_raise_errors(errors=key_errors)

        timeseries_data_series = pd.DataFrame(
            index=tskeys.index, columns=data.timesteps.values, data=timeseries_data
        ).stack()
        timeseries_data_series.index.rename("timesteps", -1, inplace=True)

        # 7) Add time dimension to the relevent DataArray and update the '='
        # dimensions with the time varying data (static data is just duplicated
        # at each timestep)
        timeseries_data_array = xr.broadcast(data[variable], data.timesteps)[0].copy()
        timeseries_data_array.loc[
            xr.DataArray.from_series(timeseries_data_series).coords
        ] = xr.DataArray.from_series(timeseries_data_series).values

        data[variable] = timeseries_data_array

    # Add timestep_resolution by looking at the time difference between timestep n
    # and timestep n + 1 for all timesteps
    time_delta = (data.timesteps.shift(timesteps=-1) - data.timesteps).to_series()

    # Last timestep has no n + 1, so will be NaT (not a time),
    # we duplicate the penultimate time_delta instead
    time_delta[-1] = time_delta[-2]
    time_delta.name = "timestep_resolution"
    # Time resolution is saved in hours (i.e. seconds / 3600)
    data["timestep_resolution"] = xr.DataArray.from_series(
        time_delta.dt.total_seconds() / 3600
    )

    data["timestep_weights"] = xr.DataArray(
        np.ones(len(data.timesteps)), dims=["timesteps"]
    )

    return data


def add_max_demand_timesteps(model_data):
    max_demand_timesteps = []

    # Get all loc_techs with a demand resource
    loc_techs_with_demand_resource = list(
        set(model_data.coords["loc_techs_finite_resource"].values).intersection(
            model_data.coords["loc_techs_demand"].values
        )
    )

    for carrier in list(model_data.carriers.data):
        # Filter demand loc_techs for this carrier
        loc_techs = [
            i
            for i in loc_techs_with_demand_resource
            if "{}::{}".format(i, carrier)
            in model_data.coords["loc_tech_carriers_con"].values
        ]

        carrier_demand = (
            model_data.resource.loc[dict(loc_techs_finite_resource=loc_techs)]
            .sum(dim="loc_techs_finite_resource")
            .copy()
        )

        # Only keep negative (=demand) values
        carrier_demand[carrier_demand.values > 0] = 0

        max_demand_timesteps.append(carrier_demand.to_series().idxmin())

    model_data["max_demand_timesteps"] = xr.DataArray(
        max_demand_timesteps, dims=["carriers"]
    )

    return model_data


def add_zero_carrier_ratio_sets(model_data):
    carrier_ratios = model_data.get("carrier_ratios", None)

    if carrier_ratios is None:
        return model_data

    zero_dims = (
        carrier_ratios.where(carrier_ratios == 0)
        .dropna("loc_tech_carriers_conversion_plus", how="all")
        .dropna("carrier_tiers", how="all")
    )

    if zero_dims.any().item() is False:
        return model_data

    zero_dims = zero_dims.stack(
        loc_tech_carrier_tiers_conversion_plus_zero_ratio=[
            "loc_tech_carriers_conversion_plus",
            "carrier_tiers",
        ]
    ).dropna("loc_tech_carrier_tiers_conversion_plus_zero_ratio", how="all")

    return model_data.assign_coords(
        loc_tech_carrier_tiers_conversion_plus_zero_ratio_constraint=[
            "::".join(i)
            for i in zero_dims.loc_tech_carrier_tiers_conversion_plus_zero_ratio.values
        ]
    )


def final_timedimension_processing(model_data):

    # Final checking of the data
    model_data, final_check_comments, warns, errors = checks.check_model_data(
        model_data
    )
    exceptions.print_warnings_and_raise_errors(warnings=warns, errors=errors)

    model_data = add_max_demand_timesteps(model_data)
    model_data = add_zero_carrier_ratio_sets(model_data)

    model_data = reorganise_xarray_dimensions(model_data)

    return model_data



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

conversion.py
~~~~~~~~~~~~~~~~~

Conversion technology constraints.

"""

import pyomo.core as po  # pylint: disable=import-error

from calliope.backend.pyomo.util import get_param

ORDER = 20  # order in which to invoke constraints relative to other constraint files


def load_constraints(backend_model):
    sets = backend_model.__calliope_model_data["sets"]

    if "loc_techs_balance_conversion_constraint" in sets:
        backend_model.balance_conversion_constraint = po.Constraint(
            backend_model.loc_techs_balance_conversion_constraint,
            backend_model.timesteps,
            rule=balance_conversion_constraint_rule,
        )

    if "loc_techs_cost_var_conversion_constraint" in sets:
        backend_model.cost_var_conversion_constraint = po.Constraint(
            backend_model.costs,
            backend_model.loc_techs_cost_var_conversion_constraint,
            backend_model.timesteps,
            rule=cost_var_conversion_constraint_rule,
        )


def balance_conversion_constraint_rule(backend_model, loc_tech, timestep):
    """
    Balance energy carrier consumption and production

    .. container:: scrolling-wrapper

        .. math::

            -1 * \\boldsymbol{carrier_{con}}(loc::tech::carrier, timestep)
            \\times \\eta_{energy}(loc::tech, timestep)
            = \\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)
            \\times \\eta_{energy}(loc::tech, timestep)
            \\quad \\forall loc::tech \\in locs::techs_{conversion},
            \\forall timestep \\in timesteps
    """
    model_data_dict = backend_model.__calliope_model_data["data"]

    loc_tech_carrier_out = model_data_dict["lookup_loc_techs_conversion"][
        ("out", loc_tech)
    ]
    loc_tech_carrier_in = model_data_dict["lookup_loc_techs_conversion"][
        ("in", loc_tech)
    ]

    energy_eff = get_param(backend_model, "energy_eff", (loc_tech, timestep))

    return (
        backend_model.carrier_prod[loc_tech_carrier_out, timestep]
        == -1 * backend_model.carrier_con[loc_tech_carrier_in, timestep] * energy_eff
    )


def cost_var_conversion_constraint_rule(backend_model, cost, loc_tech, timestep):
    """
    Add time-varying conversion technology costs

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{cost_{var}}(loc::tech, cost, timestep) =
            \\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)
            \\times timestep_{weight}(timestep) \\times cost_{om, prod}(loc::tech, cost, timestep)
            +
            \\boldsymbol{carrier_{con}}(loc::tech::carrier, timestep)
            \\times timestep_{weight}(timestep) \\times cost_{om, con}(loc::tech, cost, timestep)
            \\quad \\forall loc::tech \\in loc::techs_{cost_{var}, conversion}
    """
    model_data_dict = backend_model.__calliope_model_data
    weight = backend_model.timestep_weights[timestep]

    loc_tech_carrier_in = model_data_dict["data"]["lookup_loc_techs_conversion"][
        ("in", loc_tech)
    ]

    loc_tech_carrier_out = model_data_dict["data"]["lookup_loc_techs_conversion"][
        ("out", loc_tech)
    ]

    cost_om_prod = get_param(backend_model, "cost_om_prod", (cost, loc_tech, timestep))
    cost_om_con = get_param(backend_model, "cost_om_con", (cost, loc_tech, timestep))
    if po.value(cost_om_prod):
        cost_prod = (
            cost_om_prod
            * weight
            * backend_model.carrier_prod[loc_tech_carrier_out, timestep]
        )
    else:
        cost_prod = 0

    if po.value(cost_om_con):
        cost_con = (
            cost_om_con
            * weight
            * -1
            * backend_model.carrier_con[loc_tech_carrier_in, timestep]
        )
    else:
        cost_con = 0

    backend_model.cost_var_rhs[cost, loc_tech, timestep] = cost_prod + cost_con

    return (
        backend_model.cost_var[cost, loc_tech, timestep]
        == backend_model.cost_var_rhs[cost, loc_tech, timestep]
    )



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

constraint_sets.py
~~~~~~~~~~~~~~~~~~

loc_techs, loc_carriers, and loc_tech_carriers subsets used per constraint, to
reduce constraint complexity

"""

from itertools import product

import numpy as np

from calliope.preprocess.util import constraint_exists, concat_iterable
from calliope.preprocess import checks
from calliope import exceptions


def generate_constraint_sets(model_run):
    """
    Generate loc-tech sets for a given pre-processed ``model_run``

    Parameters
    ----------
    model_run : AttrDict
    """

    sets = model_run.sets
    ## From here on, everything is a `key=value` pair within a dictionary

    constraint_sets = dict()
    # energy_balance.py
    constraint_sets["loc_carriers_system_balance_constraint"] = sets.loc_carriers
    constraint_sets[
        "loc_techs_balance_supply_constraint"
    ] = sets.loc_techs_finite_resource_supply
    constraint_sets[
        "loc_techs_balance_demand_constraint"
    ] = sets.loc_techs_finite_resource_demand
    constraint_sets[
        "loc_techs_resource_availability_supply_plus_constraint"
    ] = sets.loc_techs_finite_resource_supply_plus
    constraint_sets[
        "loc_techs_balance_transmission_constraint"
    ] = sets.loc_techs_transmission
    constraint_sets[
        "loc_techs_balance_supply_plus_constraint"
    ] = sets.loc_techs_supply_plus
    constraint_sets["loc_techs_balance_storage_constraint"] = sets.loc_techs_storage
    if model_run.run.cyclic_storage is True:
        constraint_sets["loc_techs_storage_initial_constraint"] = [
            i
            for i in sets.loc_techs_store
            if constraint_exists(model_run, i, "constraints.storage_initial")
            is not None
        ]
    constraint_sets["loc_techs_storage_discharge_depth"] = [
        i
        for i in sets.loc_techs_store
        if constraint_exists(model_run, i, "constraints.storage_discharge_depth")
    ]
    constraint_sets["carriers_reserve_margin_constraint"] = [
        i
        for i in sets.carriers
        if i in model_run.model.get_key("reserve_margin", {}).keys()
    ]
    # clustering-specific balance constraints
    if model_run.model.get_key(
        "time.function", None
    ) == "apply_clustering" and model_run.model.get_key(
        "time.function_options.storage_inter_cluster", True
    ):
        set_name = "loc_techs_balance_storage_inter_cluster_constraint"
        constraint_sets[set_name] = sets.loc_techs_store

    # costs.py
    constraint_sets["loc_techs_cost_constraint"] = sets.loc_techs_cost
    constraint_sets[
        "loc_techs_cost_investment_constraint"
    ] = sets.loc_techs_investment_cost
    constraint_sets["loc_techs_cost_var_constraint"] = [
        i
        for i in sets.loc_techs_om_cost
        if i not in sets.loc_techs_conversion_plus + sets.loc_techs_conversion
    ]

    # export.py
    constraint_sets["loc_carriers_update_system_balance_constraint"] = [
        i
        for i in sets.loc_carriers
        if sets.loc_techs_export
        and any(
            [
                "{0}::{2}".format(*j.split("::")) == i
                for j in sets.loc_tech_carriers_export
            ]
        )
    ]
    constraint_sets[
        "loc_tech_carriers_export_balance_constraint"
    ] = sets.loc_tech_carriers_export
    constraint_sets["loc_techs_update_costs_var_constraint"] = [
        i for i in sets.loc_techs_om_cost if i in sets.loc_techs_export
    ]

    constraint_sets["loc_tech_carriers_export_max_constraint"] = [
        i
        for i in sets.loc_tech_carriers_export
        if constraint_exists(model_run, i.rsplit("::", 1)[0], "constraints.export_cap")
        is not None
    ]

    # capacity.py
    constraint_sets["loc_techs_storage_capacity_constraint"] = [
        i for i in sets.loc_techs_store if i not in sets.loc_techs_milp
    ]
    constraint_sets["loc_techs_energy_capacity_storage_constraint_old"] = [
        i
        for i in sets.loc_techs_store
        if constraint_exists(model_run, i, "constraints.charge_rate")
    ]
    constraint_sets["loc_techs_energy_capacity_storage_equals_constraint"] = [
        i
        for i in sets.loc_techs_store
        if constraint_exists(
            model_run, i, "constraints.energy_cap_per_storage_cap_equals"
        )
    ]
    constraint_sets["loc_techs_energy_capacity_storage_min_constraint"] = [
        i
        for i in sets.loc_techs_store
        if constraint_exists(model_run, i, "constraints.energy_cap_per_storage_cap_min")
        and not constraint_exists(
            model_run, i, "constraints.energy_cap_per_storage_cap_equals"
        )
    ]
    constraint_sets["loc_techs_energy_capacity_storage_max_constraint"] = [
        i
        for i in sets.loc_techs_store
        if constraint_exists(model_run, i, "constraints.energy_cap_per_storage_cap_max")
        and not constraint_exists(
            model_run, i, "constraints.energy_cap_per_storage_cap_equals"
        )
    ]
    constraint_sets["loc_techs_resource_capacity_constraint"] = [
        i
        for i in sets.loc_techs_finite_resource_supply_plus
        if any(
            [
                constraint_exists(model_run, i, "constraints.resource_cap_equals"),
                constraint_exists(model_run, i, "constraints.resource_cap_max"),
                constraint_exists(model_run, i, "constraints.resource_cap_min"),
            ]
        )
    ]
    constraint_sets["loc_techs_resource_capacity_equals_energy_capacity_constraint"] = [
        i
        for i in sets.loc_techs_finite_resource_supply_plus
        if constraint_exists(model_run, i, "constraints.resource_cap_equals_energy_cap")
    ]
    constraint_sets["loc_techs_resource_area_constraint"] = sets.loc_techs_area
    constraint_sets["loc_techs_resource_area_per_energy_capacity_constraint"] = [
        i
        for i in sets.loc_techs_area
        if constraint_exists(model_run, i, "constraints.resource_area_per_energy_cap")
        is not None
    ]
    constraint_sets["locs_resource_area_capacity_per_loc_constraint"] = [
        i
        for i in sets.locs
        if model_run.locations[i].get_key("available_area", None) is not None
        and sets.loc_techs_area
    ]
    constraint_sets["loc_techs_energy_capacity_constraint"] = [
        i
        for i in sets.loc_techs
        if i not in sets.loc_techs_milp + sets.loc_techs_purchase
    ]
    constraint_sets["techs_energy_capacity_systemwide_constraint"] = [
        i
        for i in sets.techs
        if model_run.get_key(
            "techs.{}.constraints.energy_cap_max_systemwide".format(i), None
        )
        or model_run.get_key(
            "techs.{}.constraints.energy_cap_equals_systemwide".format(i), None
        )
    ]

    # dispatch.py
    constraint_sets["loc_tech_carriers_carrier_production_max_constraint"] = [
        i
        for i in sets.loc_tech_carriers_prod
        if i not in sets.loc_tech_carriers_conversion_plus
        and i.rsplit("::", 1)[0] not in sets.loc_techs_milp
    ]
    constraint_sets["loc_tech_carriers_carrier_production_min_constraint"] = [
        i
        for i in sets.loc_tech_carriers_prod
        if i not in sets.loc_tech_carriers_conversion_plus
        and constraint_exists(
            model_run, i.rsplit("::", 1)[0], "constraints.energy_cap_min_use"
        )
        and i.rsplit("::", 1)[0] not in sets.loc_techs_milp
    ]
    constraint_sets["loc_tech_carriers_carrier_consumption_max_constraint"] = [
        i
        for i in sets.loc_tech_carriers_con
        if i.rsplit("::", 1)[0]
        in sets.loc_techs_demand + sets.loc_techs_storage + sets.loc_techs_transmission
        and i.rsplit("::", 1)[0] not in sets.loc_techs_milp
    ]
    constraint_sets["loc_techs_resource_max_constraint"] = sets.loc_techs_supply_plus
    constraint_sets["loc_tech_carriers_ramping_constraint"] = [
        i
        for i in sets.loc_tech_carriers_prod
        if i.rsplit("::", 1)[0] in sets.loc_techs_ramping
    ]
    # clustering-specific dispatch constraints
    if model_run.model.get_key(
        "time.function", None
    ) == "apply_clustering" and model_run.model.get_key(
        "time.function_options.storage_inter_cluster", True
    ):
        constraint_sets["loc_techs_storage_intra_max_constraint"] = sets.loc_techs_store
        constraint_sets["loc_techs_storage_intra_min_constraint"] = sets.loc_techs_store
        constraint_sets["loc_techs_storage_inter_max_constraint"] = sets.loc_techs_store
        constraint_sets["loc_techs_storage_inter_min_constraint"] = sets.loc_techs_store
    else:
        constraint_sets["loc_techs_storage_max_constraint"] = sets.loc_techs_store

    # milp.py
    constraint_sets["loc_techs_unit_commitment_milp_constraint"] = sets.loc_techs_milp
    constraint_sets["loc_techs_unit_capacity_milp_constraint"] = sets.loc_techs_milp
    constraint_sets["loc_tech_carriers_carrier_production_max_milp_constraint"] = [
        i
        for i in sets.loc_tech_carriers_prod
        if i not in sets.loc_tech_carriers_conversion_plus
        and i.rsplit("::", 1)[0] in sets.loc_techs_milp
    ]
    constraint_sets[
        "loc_techs_carrier_production_max_conversion_plus_milp_constraint"
    ] = [i for i in sets.loc_techs_conversion_plus if i in sets.loc_techs_milp]
    constraint_sets["loc_tech_carriers_carrier_production_min_milp_constraint"] = [
        i
        for i in sets.loc_tech_carriers_prod
        if i not in sets.loc_tech_carriers_conversion_plus
        and constraint_exists(
            model_run, i.rsplit("::", 1)[0], "constraints.energy_cap_min_use"
        )
        and i.rsplit("::", 1)[0] in sets.loc_techs_milp
    ]
    constraint_sets[
        "loc_techs_carrier_production_min_conversion_plus_milp_constraint"
    ] = [
        i
        for i in sets.loc_techs_conversion_plus
        if constraint_exists(model_run, i, "constraints.energy_cap_min_use")
        and i in sets.loc_techs_milp
    ]
    constraint_sets["loc_tech_carriers_carrier_consumption_max_milp_constraint"] = [
        i
        for i in sets.loc_tech_carriers_con
        if i.rsplit("::", 1)[0]
        in sets.loc_techs_demand + sets.loc_techs_storage + sets.loc_techs_transmission
        and i.rsplit("::", 1)[0] in sets.loc_techs_milp
    ]
    constraint_sets["loc_techs_energy_capacity_units_milp_constraint"] = [
        i
        for i in sets.loc_techs_milp
        if constraint_exists(model_run, i, "constraints.energy_cap_per_unit")
        is not None
    ]
    constraint_sets["loc_techs_storage_capacity_units_milp_constraint"] = [
        i for i in sets.loc_techs_milp if i in sets.loc_techs_store
    ]
    constraint_sets["loc_techs_energy_capacity_max_purchase_milp_constraint"] = [
        i
        for i in sets.loc_techs_purchase
        if (
            constraint_exists(model_run, i, "constraints.energy_cap_equals") is not None
            or (
                constraint_exists(model_run, i, "constraints.energy_cap_max")
                is not None
                and constraint_exists(model_run, i, "constraints.energy_cap_max")
                != np.inf
            )
        )
    ]
    constraint_sets["loc_techs_energy_capacity_min_purchase_milp_constraint"] = [
        i
        for i in sets.loc_techs_purchase
        if (
            not constraint_exists(model_run, i, "constraints.energy_cap_equals")
            and constraint_exists(model_run, i, "constraints.energy_cap_min")
        )
    ]
    constraint_sets["loc_techs_storage_capacity_max_purchase_milp_constraint"] = [
        i
        for i in set(sets.loc_techs_purchase).intersection(sets.loc_techs_store)
        if (
            constraint_exists(model_run, i, "constraints.storage_cap_equals")
            is not None
            or (
                constraint_exists(model_run, i, "constraints.storage_cap_max")
                is not None
                and constraint_exists(model_run, i, "constraints.storage_cap_max")
                != np.inf
            )
        )
    ]
    constraint_sets["loc_techs_storage_capacity_min_purchase_milp_constraint"] = [
        i
        for i in set(sets.loc_techs_purchase).intersection(sets.loc_techs_store)
        if (
            not constraint_exists(model_run, i, "constraints.storage_cap_equals")
            and constraint_exists(model_run, i, "constraints.storage_cap_min")
        )
    ]
    constraint_sets["loc_techs_update_costs_investment_units_milp_constraint"] = [
        i
        for i in sets.loc_techs_milp
        if i in sets.loc_techs_investment_cost
        and any(
            constraint_exists(model_run, i, "costs.{}.purchase".format(j))
            for j in model_run.sets.costs
        )
    ]

    # loc_techs_purchase technologies only exist because they have defined a purchase cost
    constraint_sets[
        "loc_techs_update_costs_investment_purchase_milp_constraint"
    ] = sets.loc_techs_purchase

    constraint_sets["techs_unit_capacity_systemwide_milp_constraint"] = [
        i
        for i in sets.techs
        if model_run.get_key(
            "techs.{}.constraints.units_max_systemwide".format(i), None
        )
        or model_run.get_key(
            "techs.{}.constraints.units_equals_systemwide".format(i), None
        )
    ]
    constraint_sets[
        "loc_techs_asynchronous_prod_con_milp_constraint"
    ] = sets.loc_techs_asynchronous_prod_con

    # conversion.py
    constraint_sets[
        "loc_techs_balance_conversion_constraint"
    ] = sets.loc_techs_conversion
    constraint_sets[
        "loc_techs_cost_var_conversion_constraint"
    ] = sets.loc_techs_om_cost_conversion

    # conversion_plus.py
    constraint_sets[
        "loc_techs_balance_conversion_plus_primary_constraint"
    ] = sets.loc_techs_conversion_plus
    constraint_sets["loc_techs_carrier_production_max_conversion_plus_constraint"] = [
        i for i in sets.loc_techs_conversion_plus if i not in sets.loc_techs_milp
    ]
    constraint_sets["loc_techs_carrier_production_min_conversion_plus_constraint"] = [
        i
        for i in sets.loc_techs_conversion_plus
        if constraint_exists(model_run, i, "constraints.energy_cap_min_use")
        and i not in sets.loc_techs_milp
    ]
    constraint_sets[
        "loc_techs_cost_var_conversion_plus_constraint"
    ] = sets.loc_techs_om_cost_conversion_plus
    constraint_sets[
        "loc_techs_balance_conversion_plus_in_2_constraint"
    ] = sets.loc_techs_in_2
    constraint_sets[
        "loc_techs_balance_conversion_plus_in_3_constraint"
    ] = sets.loc_techs_in_3
    constraint_sets[
        "loc_techs_balance_conversion_plus_out_2_constraint"
    ] = sets.loc_techs_out_2
    constraint_sets[
        "loc_techs_balance_conversion_plus_out_3_constraint"
    ] = sets.loc_techs_out_3

    # network.py
    constraint_sets[
        "loc_techs_symmetric_transmission_constraint"
    ] = sets.loc_techs_transmission

    # policy.py
    for sense in ["min", "max", "equals"]:
        constraint_sets[f"techlists_group_share_energy_cap_{sense}_constraint"] = [
            i
            for i in sets.techlists
            if f"energy_cap_{sense}"
            in model_run.model.get_key("group_share.{}".format(i), {}).keys()
        ]
        constraint_sets[
            f"techlists_carrier_group_share_carrier_prod_{sense}_constraint"
        ] = [
            i + "::" + carrier
            for i in sets.techlists
            if f"carrier_prod_{sense}"
            in model_run.model.get_key("group_share.{}".format(i), {}).keys()
            for carrier in sets.carriers
            if carrier
            in model_run.model.get_key(
                f"group_share.{i}.carrier_prod_{sense}", {}
            ).keys()
        ]

    # group.py
    group_constraints = {
        name: data
        for name, data in model_run["group_constraints"].items()
        if data.get("exists", True)
    }
    constraint_sets["group_constraints"] = set()
    for group_constraint_name, group_constraint in group_constraints.items():
        tech_groups = [
            [
                k
                for k, v in checks.DEFAULTS.tech_groups.items()
                if i in v["allowed_group_constraints"]
            ]
            for i in group_constraint.keys()
            if i not in ["techs", "locs", "exists"]
        ]
        allowed_tech_groups = set(tech_groups[0]).intersection(*tech_groups)
        allowed_techs = sum(
            [sets["techs_{}".format(i)] for i in allowed_tech_groups], []
        )
        techs = group_constraint.get("techs", allowed_techs)

        locs = group_constraint.get("locs", sets["locs"])

        # If there are transmission techs, keep only those that link to allowed locations
        techs = [i for i in techs if ":" not in techs or i.split(":")[-1] in locs]

        trans_techs = set(techs).intersection(sets["techs_transmission_names"])
        for i in trans_techs:
            techs += [i + ":" + j for j in locs]
            techs.remove(i)

        # If the group constraint defines its own techs, remove those that are not allowed
        techs = list(set(techs).intersection(allowed_techs))

        # All possible loc_techs for this constraint
        loc_techs_all = list(
            set(concat_iterable([(l, t) for l, t in product(locs, techs)], ["::"]))
        )

        # Some loc_techs may not actually exist in the actual model,
        # so we must filter with actually exising loc_techs
        loc_techs = [i for i in loc_techs_all if i in sets.loc_techs]
        default_group_config = checks.DEFAULTS.group_constraints.default_group
        _constraints = {
            k: v
            for k, v in group_constraint.items()
            if k not in ["locs", "techs", "exists"]
        }
        if any(
            isinstance(default_group_config.get(_constraint, False), dict)
            and "default_carrier" in default_group_config[_constraint].keys()
            for _constraint in _constraints.keys()
        ):
            if len(_constraints) > 1:
                raise exceptions.ModelError(
                    "Can only handle one constraint in a group constraint if one of them is carrier-based"
                )
            _name, _config = list(_constraints.items())[0]
            loc_tech_carrier_dict = _get_carrier_group_constraint_loc_techs(
                loc_techs, locs, _config, _name, sets, constraint_sets
            )
            if any(len(val) == 0 for val in loc_tech_carrier_dict.values()):
                exceptions.warn(
                    f"Constraint group `{group_constraint_name}` will be completely ignored since there are no valid location::technology::carrier combinations"
                )
                continue
            else:
                for key, loc_tech_carriers in loc_tech_carrier_dict.items():
                    constraint_sets[
                        key.format(group_constraint_name)
                    ] = loc_tech_carriers

        else:
            if len(loc_techs) == 0:
                exceptions.warn(
                    f"Constraint group `{group_constraint_name}` will be completely ignored since there are no valid location::technology combinations"
                )
                break
            constraint_sets[
                "group_constraint_loc_techs_{}".format(group_constraint_name)
            ] = loc_techs
        _add_to_group_constraint_mapping(
            constraint_sets, group_constraint_name, list(_constraints.keys())
        )
    constraint_sets["group_constraints"] = list(constraint_sets["group_constraints"])
    return constraint_sets


def _get_carrier_group_constraint_loc_techs(
    loc_techs, locs, config, constraint_name, sets, constraint_sets
):
    flow = "con" if "_con_" in constraint_name else "prod"
    if len(config.keys()) > 1:
        raise exceptions.ModelError(
            "Can only handle one carrier per group constraint that is carrier-based"
        )
    carrier = list(config.keys())[0]
    if "net_import" in constraint_name:
        _loc_tech_carriers = _get_net_import_loc_tech_carrier_subset(
            loc_techs,
            sets["loc_tech_carriers_con"] + sets["loc_tech_carriers_prod"],
            carrier,
            locs,
        )
    else:
        _loc_tech_carriers = _get_loc_tech_carrier_subset(
            loc_techs, sets[f"loc_tech_carriers_{flow}"], carrier, locs
        )

    if "share" in constraint_name:
        lhs_loc_tech_carriers = _loc_tech_carriers
        if "demand" in constraint_name or "import" in constraint_name:
            rhs_loc_tech_carriers = _get_loc_tech_carrier_subset(
                sets["loc_techs_demand"], sets["loc_tech_carriers_con"], carrier, locs
            )
        elif flow == "con":
            rhs_loc_tech_carriers = _get_loc_tech_carrier_subset(
                sets["loc_techs_demand"] + sets["loc_techs_conversion_all"],
                sets["loc_tech_carriers_con"],
                carrier,
                locs,
            )
        elif flow == "prod":
            rhs_loc_tech_carriers = _get_loc_tech_carrier_subset(
                sets["loc_techs_supply_conversion_all"],
                sets["loc_tech_carriers_prod"],
                carrier,
                locs,
            )
        return {
            "group_constraint_loc_tech_carriers_{}_lhs": list(lhs_loc_tech_carriers),
            "group_constraint_loc_tech_carriers_{}_rhs": list(rhs_loc_tech_carriers),
        }
    else:
        return {"group_constraint_loc_tech_carriers_{}": list(_loc_tech_carriers)}


def _get_loc_tech_carrier_subset(loc_techs, loc_tech_carriers, carrier, locs):
    return set(
        f"{loc_tech}::{carrier}"
        for loc_tech in loc_techs
        if f"{loc_tech}::{carrier}" in loc_tech_carriers
        and loc_tech.split("::")[0] in locs
    )


def _get_net_import_loc_tech_carrier_subset(
    loc_techs, loc_tech_carriers, carrier, locs
):
    return set(
        f"{loc_tech}::{carrier}"
        for loc_tech in loc_techs
        if f"{loc_tech}::{carrier}" in loc_tech_carriers
        and loc_tech.split(":")[-1] not in locs
    )


def _add_to_group_constraint_mapping(constraint_sets, group_name, constraint_names):
    for constraint_name in constraint_names:
        if (
            f"group_names_{constraint_name}" in constraint_sets.keys()
            and group_name not in constraint_sets[f"group_names_{constraint_name}"]
        ):
            constraint_sets[f"group_names_{constraint_name}"] += [group_name]
        else:
            constraint_sets[f"group_names_{constraint_name}"] = [group_name]
        constraint_sets["group_constraints"].update([constraint_name])



"""
Copyright (C) 2013-2019 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

util.py
~~~~~~~~

Plotting util functions.

"""

import numpy as np
from IPython import get_ipython


def type_of_script():
    if get_ipython() is not None:
        ipy_str = str(type(get_ipython()))
        if "zmqshell" in ipy_str:
            return "jupyter"
        if "terminal" in ipy_str:
            return "ipython"
    else:
        return "terminal"


def get_data_layout(
    get_var_data,
    get_var_layout,
    relevant_vars,
    layout,
    model,
    dataset,
    subset,
    sum_dims,
    squeeze,
    get_var_data_kwargs={},
    get_var_layout_kwargs={},
):
    """
    For each dropdown dataset in capacity and timeseries plotting, build the
    plotly data dictionary and the plotly layout dictionary.

    """
    # data_len is used to populate visibility of traces, for dropdown
    data_len = [0]
    data = []
    buttons = []
    # fill trace data and add number of traces per var to 'data_len' for use with
    # visibility. first var in loop has visibility == True by default
    visible = True

    for var in relevant_vars:
        data += get_var_data(
            var,
            model,
            dataset,
            visible,
            subset,
            sum_dims,
            squeeze,
            **get_var_data_kwargs,
        )
        data_len.append(len(data))
        visible = False

    # Initialise all visibility to False for dropdown updates
    total_data_arrays = np.array([False for i in range(data_len[-1])])
    var_num = 0
    for var in relevant_vars:
        # update visibility to True for all traces linked to this variable `var`
        visible_data = total_data_arrays.copy()
        visible_data[data_len[var_num] : data_len[var_num + 1]] = True

        # Get variable-specific layout
        var_layout = get_var_layout(var, dataset, **get_var_layout_kwargs)

        if var_num == 0:
            layout["title"] = var_layout["title"]
            if "barmode" in var_layout:
                layout["barmode"] = var_layout["barmode"]

        if len(relevant_vars) > 1:
            var_layout = [{"visible": list(visible_data)}, var_layout]
            buttons.append(dict(label=var, method="update", args=var_layout))

        var_num += 1

    # If there are multiple vars to plot, use dropdowns via 'updatemenus'
    if len(relevant_vars) > 1:
        updatemenus = list(
            [
                dict(
                    active=0,
                    buttons=buttons,
                    type="dropdown",
                    xanchor="left",
                    x=0,
                    y=1.13,
                    pad=dict(t=0.05, b=0.05, l=0.05, r=0.05),
                )
            ]
        )
        layout["updatemenus"] = updatemenus
    else:
        layout.update(var_layout)
    return data, layout


def hex_to_rgba(hex_color, opacity):
    """Embed opacity in a colour by converting calliope HEX colours to an RGBA"""
    _NUMERALS = "0123456789abcdefABCDEF"
    _HEXDEC = {v: int(v, 16) for v in (x + y for x in _NUMERALS for y in _NUMERALS)}
    hex_color = hex_color.lstrip("#")
    rgb = [_HEXDEC[hex_color[0:2]], _HEXDEC[hex_color[2:4]], _HEXDEC[hex_color[4:6]]]
    return "rgba({1}, {2}, {3}, {0})".format(opacity, *rgb)


def break_name(name, length):
    """Take a long technology name string and break it across multiple lines"""

    name_breaks = len(name) / length
    if name_breaks > 1:
        initial_name = name
        break_buffer = 0
        for name_break in range(1, int(name_breaks) + 1):
            # preferably break at a space
            breakpoint = initial_name.rfind(
                " ", name_break * length - 10, name_break * length
            )
            # -1 means rfind failed to find any space
            if breakpoint != -1:
                breakpoint += break_buffer
                name = name[:breakpoint].rstrip() + "<br>" + name[breakpoint:].lstrip()
                break_buffer += 4
            else:
                breakpoint = length * name_break + break_buffer
                name = (
                    name[:breakpoint].rstrip() + "...<br>" + name[breakpoint:].lstrip()
                )
                # '...' is len 3
                break_buffer += 7

    return name


def get_clustered_layout(dataset):
    layout = {}
    timestep_cluster = dataset.timestep_cluster.to_pandas()
    clusters = timestep_cluster.groupby(timestep_cluster).groups
    layout["xaxis"] = {}
    layout["xaxis"]["type"] = "category"
    layout["xaxis"]["tickvals"] = [
        (2 * (k - min(clusters.keys())) + 1) * len(v) / 2 - 0.5
        for k, v in clusters.items()
    ]
    layout["xaxis"]["ticktext"] = [k for k in clusters.keys()]
    layout["xaxis"]["title"] = "Clusters"

    # Make rectangles to fit in the background over every other cluster,
    # to distinguish them
    layout["shapes"] = []
    shape_template = {
        "type": "rect",
        "xref": "x",
        "yref": "paper",
        "y0": 0,
        "y1": 1,
        "line": {"width": 0},
        "layer": "below",
    }

    for cluster in clusters.keys():
        x0 = clusters[cluster][0]
        x1 = clusters[cluster][-1]
        opacity = 0.3 * (cluster % 2)
        day_shape = {"x0": x0, "x1": x1, "fillcolor": "grey", "opacity": opacity}

        shape_template.update(day_shape)
        layout["shapes"].append(shape_template.copy())

    return layout


def get_range(coordinates, axis, buffer):
    """
    Based on lat/lon of the underlying data, get a range which defines the
    bounding box for the map

    Parameters
    ----------
    coordinates: xarray DataArray
        Coordinate data of the underlying data
    axis: str
        axis of the data (lat/lon) by which to slice the coordinates DataArray
    buffer: float
        value, expressed as a fraction of the range of the underlying data,
        to add as a buffer to the bounding box

    Returns
    -------
    list, defining the range as input_range +/- input_range * buffer
    """
    _range = [
        coordinates.loc[dict(coordinates=axis)].min().item(),
        coordinates.loc[dict(coordinates=axis)].max().item(),
    ]
    _offset = abs(_range[1] - _range[0]) * 0.1
    return [_range[0] - _offset, _range[1] + _offset]



import pytest  # pylint: disable=unused-import
import calliope
import logging
import datetime
import os
import tempfile

import xarray as xr
import pandas as pd

from calliope.core.util import dataset, observed_dict

from calliope.core.util.tools import memoize, memoize_instancemethod

from calliope import exceptions
from calliope.core.util.logging import log_time
from calliope.core.util.generate_runs import generate_runs
from calliope.test.common.util import (
    python36_or_higher,
    check_error_or_warning,
    build_test_model,
)

_MODEL_NATIONAL = os.path.join(
    os.path.dirname(__file__), "..", "example_models", "national_scale", "model.yaml"
)

_MODEL_URBAN = os.path.join(
    os.path.dirname(__file__), "..", "example_models", "urban_scale", "model.yaml"
)


class TestDataset:
    @pytest.fixture()
    def loc_techs(self):
        loc_techs = [
            "region1-3::csp",
            "region1::demand_power",
            "region1-1::csp",
            "region1-2::csp",
            "region2::demand_power",
            "region1-1::demand_power",
        ]
        return loc_techs

    @pytest.fixture()
    def example_dataarray(self):
        return xr.DataArray(
            [[[0], [1], [2]], [[3], [4], [5]]],
            dims=("timesteps", "loc_techs_bar", "costs"),
            coords={
                "timesteps": ["foo", "bar"],
                "loc_techs_bar": ["1::foo", "2::bar", "3::baz"],
                "costs": ["foo"],
            },
        )

    @pytest.fixture()
    def example_one_dim_dataarray(self):
        return xr.DataArray(
            [0, 1, 2], dims=("timesteps"), coords={"timesteps": ["foo", "bar", "baz"]}
        )

    @pytest.fixture()
    def example_dataset(self, example_dataarray):
        return xr.Dataset(
            {"foo": example_dataarray, "bar": example_dataarray.squeeze()}
        )

    def test_get_loc_techs_tech(self, loc_techs):
        loc_techs = dataset.get_loc_techs(loc_techs, tech="csp")
        assert loc_techs == ["region1-3::csp", "region1-1::csp", "region1-2::csp"]

    def test_get_loc_techs_loc(self, loc_techs):
        loc_techs = dataset.get_loc_techs(loc_techs, loc="region1-1")
        assert loc_techs == ["region1-1::csp", "region1-1::demand_power"]

    def test_get_loc_techs_loc_and_tech(self, loc_techs):
        loc_techs = dataset.get_loc_techs(loc_techs, tech="demand_power", loc="region1")
        assert loc_techs == ["region1::demand_power"]

    def test_split_loc_tech_to_dataarray(self, example_dataarray):
        formatted_array = dataset.split_loc_techs(example_dataarray)
        assert isinstance(formatted_array, xr.DataArray)
        assert formatted_array.dims == ("costs", "locs", "techs", "timesteps")

    def test_split_loc_tech_to_series(self, example_dataarray):
        formatted_series = dataset.split_loc_techs(
            example_dataarray, return_as="Series"
        )
        assert isinstance(formatted_series, pd.Series)
        assert formatted_series.index.names == ["costs", "locs", "techs", "timesteps"]

    def test_split_loc_tech_to_multiindex_dataarray(self, example_dataarray):
        formatted_array = dataset.split_loc_techs(
            example_dataarray, return_as="MultiIndex DataArray"
        )
        assert isinstance(formatted_array, xr.DataArray)
        assert formatted_array.dims == ("timesteps", "loc_techs_bar", "costs")
        assert isinstance(formatted_array.loc_techs_bar.to_index(), pd.MultiIndex)

    def test_split_loc_tech_too_many_loc_tech_dims(self, example_dataarray):
        _array = example_dataarray.rename({"costs": "loc_techs_2"})
        with pytest.raises(exceptions.ModelError) as excinfo:
            dataset.split_loc_techs(_array)
        assert check_error_or_warning(
            excinfo, "Cannot split loc_techs or loc_tech_carriers dimension"
        )

    def test_split_loc_tech_one_dim_to_dataarray(self, example_one_dim_dataarray):
        formatted_array = dataset.split_loc_techs(example_one_dim_dataarray)
        assert isinstance(formatted_array, xr.DataArray)
        assert formatted_array.dims == ("timesteps",)

    def test_split_loc_tech_one_dim_to_series(self, example_one_dim_dataarray):
        formatted_series = dataset.split_loc_techs(
            example_one_dim_dataarray, return_as="Series"
        )
        assert isinstance(formatted_series, pd.Series)
        assert formatted_series.index.names == ["timesteps"]

    def test_split_loc_tech_one_dim_to_multiindex_dataarray(
        self, example_one_dim_dataarray
    ):
        formatted_array = dataset.split_loc_techs(
            example_one_dim_dataarray, return_as="MultiIndex DataArray"
        )
        assert isinstance(formatted_array, xr.DataArray)
        assert formatted_array.dims == ("timesteps",)

    def test_split_loc_tech_unknown_output(
        self, example_dataarray, example_one_dim_dataarray
    ):
        for array in [example_dataarray, example_one_dim_dataarray]:
            with pytest.raises(ValueError) as excinfo:
                dataset.split_loc_techs(array, return_as="foo")
            assert check_error_or_warning(
                excinfo,
                "`return_as` must be `DataArray`, `Series`, or `MultiIndex DataArray`",
            )

    def test_reorganise_dataset_dimensions(self, example_dataset):
        reorganised_dataset = dataset.reorganise_xarray_dimensions(example_dataset)
        dataset_dims = [i for i in reorganised_dataset.dims.keys()]
        assert dataset_dims == ["costs", "loc_techs_bar", "timesteps"]

    def test_reorganise_dataarray_dimensions(self, example_dataarray):
        reorganised_dataset = dataset.reorganise_xarray_dimensions(example_dataarray)
        assert reorganised_dataset.dims == ("costs", "loc_techs_bar", "timesteps")

    def test_fail_reorganise_dimensions(self):
        with pytest.raises(TypeError) as excinfo:
            dataset.reorganise_xarray_dimensions(
                ["timesteps", "loc_techs_bar", "costs"]
            )
        assert check_error_or_warning(
            excinfo, "Must provide either xarray Dataset or DataArray to be reorganised"
        )


class TestMemoization:
    @memoize_instancemethod
    def instance_method(self, a, b):
        return a + b

    def test_memoize_one_arg(self):
        @memoize
        def test(a):
            return a + 1

        assert test(1) == 2
        assert test(1) == 2

    def test_memoize_two_args(self):
        @memoize
        def test(a, b):
            return a + b

        assert test(1, 2) == 3
        assert test(1, 2) == 3

    def test_memoize_instancemethod(self):
        assert self.instance_method(1, 2) == 3
        assert self.instance_method(1, 2) == 3


class TestLogging:
    def test_set_log_verbosity(self):
        calliope.set_log_verbosity("CRITICAL", include_solver_output=True)

        assert logging.getLogger("calliope").getEffectiveLevel() == 50
        assert logging.getLogger("py.warnings").getEffectiveLevel() == 50
        assert (
            logging.getLogger("calliope.backend.pyomo.model").getEffectiveLevel() == 10
        )

        calliope.set_log_verbosity("CRITICAL", include_solver_output=False)

        assert logging.getLogger("calliope").getEffectiveLevel() == 50
        assert logging.getLogger("py.warnings").getEffectiveLevel() == 50
        assert (
            logging.getLogger("calliope.backend.pyomo.model").getEffectiveLevel() == 50
        )

        calliope.set_log_verbosity()

        assert logging.getLogger("calliope").getEffectiveLevel() == 20
        assert logging.getLogger("py.warnings").getEffectiveLevel() == 20
        assert (
            logging.getLogger("calliope.backend.pyomo.model").getEffectiveLevel() == 10
        )

    def test_timing_log(self):
        timings = {"model_creation": datetime.datetime.now()}
        logger = logging.getLogger("calliope.testlogger")

        # TODO: capture logging output and check that comment is in string
        log_time(logger, timings, "test", comment="test_comment", level="info")
        assert isinstance(timings["test"], datetime.datetime)

        log_time(logger, timings, "test2", comment=None, level="info")
        assert isinstance(timings["test2"], datetime.datetime)

        # TODO: capture logging output and check that time_since_run_start is in the string
        log_time(
            logger,
            timings,
            "test",
            comment=None,
            level="info",
            time_since_run_start=True,
        )


class TestGenerateRuns:
    @python36_or_higher
    def test_generate_runs_scenarios(self):
        runs = generate_runs(
            _MODEL_NATIONAL, scenarios="time_resampling;profiling;time_clustering"
        )
        assert len(runs) == 3
        assert runs[0].endswith(
            "--scenario time_resampling --save_netcdf out_1_time_resampling.nc --save_plots plots_1_time_resampling.html"
        )

    @python36_or_higher
    def test_generate_runs_scenarios_none_with_scenarios(self):
        runs = generate_runs(_MODEL_NATIONAL, scenarios=None)
        assert len(runs) == 2
        assert runs[0].endswith(
            "--scenario cold_fusion_with_production_share --save_netcdf out_1_cold_fusion_with_production_share.nc --save_plots plots_1_cold_fusion_with_production_share.html"
        )

    @python36_or_higher
    def test_generate_runs_scenarios_none_with_overrides(self):
        runs = generate_runs(
            _MODEL_URBAN,
            scenarios=None,
        )
        assert len(runs) == 4
        assert runs[0].endswith(
            "--scenario milp --save_netcdf out_1_milp.nc --save_plots plots_1_milp.html"
        )


class TestPandasExport:
    @pytest.fixture(scope="module")
    def model(self):
        return calliope.examples.national_scale()

    @pytest.mark.parametrize(
        "variable_name",
        sorted(
            [i for i in calliope.examples.national_scale()._model_data.data_vars.keys()]
        ),
    )
    def test_data_variables_can_be_exported_to_pandas(self, model, variable_name):
        model.get_formatted_array(variable_name).to_dataframe()


class TestObservedDict:
    def as_yaml(self, _dict, strip=False):
        if strip is True:
            _dict = {
                k: v
                for k, v in _dict.items()
                if (not isinstance(v, dict) and v is not None)
                or (isinstance(v, dict) and len(v.keys()) > 0)
            }
        return calliope.AttrDict.to_yaml(calliope.AttrDict(_dict))

    @pytest.fixture(scope="module")
    def model(self):
        return calliope.examples.national_scale()

    @pytest.fixture(scope="module")
    def observer(self):
        return xr.Dataset()

    @pytest.fixture(scope="module")
    def observed_from_dict(self, observer):
        initial_dict = {"foo": "bar", "foobar": {"baz": "fob"}}
        return observed_dict.UpdateObserverDict(
            initial_dict=initial_dict, name="test", observer=observer
        )

    @pytest.fixture(scope="module")
    def observed_from_string(self, observer):
        initial_dict = {"foo": "bar", "foobar": {"baz": "fob"}}
        return observed_dict.UpdateObserverDict(
            initial_yaml_string=self.as_yaml(initial_dict),
            name="test_2",
            observer=observer,
        )

    def test_initialise_observer(
        self, observer, observed_from_dict, observed_from_string
    ):
        assert "test" in observer.attrs.keys()
        assert "test_2" in observer.attrs.keys()
        assert observer.attrs["test"] == self.as_yaml(
            {"foo": "bar", "foobar": {"baz": "fob"}}
        )
        observer.attrs["test"] == observer.attrs["test_2"]

    def test_value_error_on_initialising(self):
        initial_dict = {"foo": "bar", "foobar": {"baz": "fob"}}
        initial_string = self.as_yaml(initial_dict)
        with pytest.raises(ValueError) as error:
            observed_dict.UpdateObserverDict(name="test_2", observer=xr.Dataset())
        assert check_error_or_warning(
            error,
            "must supply one, and only one, of initial_dict or initial_yaml_string",
        )
        with pytest.raises(ValueError) as error:
            observed_dict.UpdateObserverDict(
                initial_yaml_string=initial_string,
                initial_dict=initial_dict,
                name="test_2",
                observer=xr.Dataset(),
            )
        assert check_error_or_warning(
            error,
            "must supply one, and only one, of initial_dict or initial_yaml_string",
        )

    @pytest.mark.parametrize(
        "key1,key2,value,result",
        (
            ("foo", None, 1, {"foo": 1, "foobar": {"baz": "fob"}}),
            ("foobar", "baz", 2, {"foo": 1, "foobar": {"baz": 2}}),
            (
                "foo",
                None,
                {"baz": "fob"},
                {"foo": {"baz": "fob"}, "foobar": {"baz": 2}},
            ),
            ("foo", "baz", 3, {"foo": {"baz": 3}, "foobar": {"baz": 2}}),
            ("foo", None, {}, {"foobar": {"baz": 2}}),
            ("foo", None, 5, {"foo": 5, "foobar": {"baz": 2}}),
            ("foo", None, None, {"foobar": {"baz": 2}}),
        ),
    )
    def test_set_item_observer(
        self, observed_from_dict, observer, key1, key2, value, result
    ):
        if key2 is None:
            observed_from_dict[key1] = value
        else:
            observed_from_dict[key1][key2] = value

        assert observer.attrs["test"] == self.as_yaml(result)

    def test_update_observer(self, observed_from_dict, observer):
        observed_from_dict.update({"baz": 4})
        assert observer.attrs["test"] == self.as_yaml({"foobar": {"baz": 2}, "baz": 4})

    def test_reinstate_observer(self, observed_from_dict, observer):
        observer.attrs["test"] = "{}"
        assert observer.attrs["test"] == "{}"
        observed_from_dict["foo"] = 5
        assert observer.attrs["test"] == self.as_yaml(
            {"foo": 5, "foobar": {"baz": 2}, "baz": 4}
        )

    def test_model_config(self, model):
        assert hasattr(model, "model_config")
        assert "model_config" in model._model_data.attrs.keys()
        assert model._model_data.attrs["model_config"] == self.as_yaml(
            model.model_config, strip=True
        )

        model.model_config["name"] = "new name"
        assert model.model_config["name"] == "new name"
        assert model._model_data.attrs["model_config"] == self.as_yaml(
            model.model_config, strip=True
        )

    def test_run_config(self, model):
        assert hasattr(model, "run_config")
        assert "run_config" in model._model_data.attrs.keys()
        assert model._model_data.attrs["run_config"] == self.as_yaml(
            model.run_config, strip=True
        )

        model.run_config["solver"] = "cplex"
        assert model.run_config["solver"] == "cplex"
        assert model._model_data.attrs["run_config"] == self.as_yaml(
            model.run_config, strip=True
        )

    def test_load_from_netcdf(self, model):

        with tempfile.TemporaryDirectory() as tempdir:
            out_path = os.path.join(tempdir, "model.nc")
            model.to_netcdf(out_path)

            model_from_disk = calliope.read_netcdf(out_path)
            assert hasattr(model_from_disk, "run_config")
            assert "run_config" in model_from_disk._model_data.attrs.keys()
            assert hasattr(model_from_disk, "model_config")
            assert "model_config" in model_from_disk._model_data.attrs.keys()



import os
import tempfile

import pytest  # pylint: disable=unused-import

import calliope
from calliope import exceptions
from calliope.test.common.util import check_error_or_warning


class TestIO:
    @pytest.fixture(scope="module")
    def model(self):
        model = calliope.examples.national_scale()
        model.run()
        return model

    def test_save_netcdf(self, model):
        bool_attrs = [
            k for k, v in model._model_data.attrs.items() if isinstance(v, bool)
        ]

        with tempfile.TemporaryDirectory() as tempdir:
            out_path = os.path.join(tempdir, "model.nc")
            model.to_netcdf(out_path)
            assert os.path.isfile(out_path)

        # Ensure that boolean attrs have not changed
        for k in bool_attrs:
            assert isinstance(model._model_data.attrs[k], bool)

    def test_save_csv_dir_mustnt_exist(self, model):
        with tempfile.TemporaryDirectory() as tempdir:
            out_path = os.path.join(tempdir)
            with pytest.raises(FileExistsError):
                model.to_csv(out_path)

    @pytest.mark.parametrize(
        "file_name",
        sorted(
            [
                "inputs_{}.csv".format(i)
                for i in calliope.examples.national_scale().inputs.data_vars.keys()
            ]
        ),
    )
    def test_save_csv(self, model, file_name):
        with tempfile.TemporaryDirectory() as tempdir:
            out_path = os.path.join(tempdir, "out_dir")
            model.to_csv(out_path)
            assert os.path.isfile(os.path.join(out_path, file_name))

            with open(
                os.path.join(out_path, "inputs_energy_cap_max_systemwide.csv"), "r"
            ) as f:
                assert "demand_power" not in f.read()

    def test_save_csv_no_dropna(self, model):
        with tempfile.TemporaryDirectory() as tempdir:
            out_path = os.path.join(tempdir, "out_dir")
            model.to_csv(out_path, dropna=False)

            with open(
                os.path.join(out_path, "inputs_energy_cap_max_systemwide.csv"), "r"
            ) as f:
                assert "demand_power" in f.read()

    def test_save_csv_not_optimal(self):
        model = calliope.examples.national_scale(
            scenario="check_feasibility", override_dict={"run.cyclic_storage": False}
        )

        model.run()

        with tempfile.TemporaryDirectory() as tempdir:
            out_path = os.path.join(tempdir, "out_dir")
            with pytest.warns(exceptions.ModelWarning):
                model.to_csv(out_path, dropna=False)

    def test_solve_save_read_netcdf(self, model):
        with tempfile.TemporaryDirectory() as tempdir:
            out_path = os.path.join(tempdir, "model.nc")
            model.to_netcdf(out_path)
            assert os.path.isfile(out_path)

            model_from_disk = calliope.read_netcdf(out_path)
            # FIXME test for some data in model_from_disk

    def test_save_read_solve_save_netcdf(self, model):

        with tempfile.TemporaryDirectory() as tempdir:
            out_path = os.path.join(tempdir, "model.nc")
            model.to_netcdf(out_path)
            model_from_disk = calliope.read_netcdf(out_path)

        # Ensure _model_run doesn't exist to simulate a re-run
        # via the backend
        delattr(model_from_disk, "_model_run")
        model_from_disk.run(force_rerun=True)
        assert not hasattr(model_from_disk, "_model_run")

        with tempfile.TemporaryDirectory() as tempdir:
            out_path = os.path.join(tempdir, "model.nc")
            model_from_disk.to_netcdf(out_path)
            assert os.path.isfile(out_path)

    def test_netcdf_roundtrip_modelrun_to_yaml(self, model):
        with tempfile.TemporaryDirectory() as tempdir:
            out_path_nc = os.path.join(tempdir, "model.nc")
            out_path_yaml = os.path.join(tempdir, "modelrun.yaml")
            model.to_netcdf(out_path_nc)
            model_from_disk = calliope.read_netcdf(out_path_nc)

            model_from_disk.save_commented_model_yaml(out_path_yaml)
            assert os.path.isfile(out_path_yaml)

    def test_netcdf_to_yaml_raises_if_attrs_missing(self, model):
        with tempfile.TemporaryDirectory() as tempdir:
            out_path_yaml = os.path.join(tempdir, "modelrun.yaml")
            model._model_run = {}
            with pytest.raises(KeyError) as error:
                model.save_commented_model_yaml(out_path_yaml)

        assert check_error_or_warning(
            error, "This model does not have the fully built model attached"
        )

    def test_save_lp(self, model):
        with tempfile.TemporaryDirectory() as tempdir:
            out_path = os.path.join(tempdir, "model.lp")
            model.to_lp(out_path)

            with open(out_path, "r") as f:
                assert "\nmin \nobj:\n+1 cost(monetary__region1_1__csp_)" in f.read()

    def test_save_per_spore(self):
        with tempfile.TemporaryDirectory() as tempdir:
            os.mkdir(os.path.join(tempdir, "output"))
            model = calliope.examples.national_scale(
                scenario="spores",
                override_dict={
                    "run.spores_options.save_per_spore": True,
                    "run.spores_options.save_per_spore_path": os.path.join(
                        tempdir, "output/spore_{}.nc"
                    ),
                },
            )
            model.run()

            for i in ["0", "1", "2", "3"]:
                assert os.path.isfile(os.path.join(tempdir, "output", f"spore_{i}.nc"))
            assert not os.path.isfile(os.path.join(tempdir, "output.nc"))



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

group.py
~~~~~~~~

Group constraints.

"""

import logging

import numpy as np
import pyomo.core as po  # pylint: disable=import-error

from calliope.backend.pyomo.util import (
    loc_tech_is_in,
    get_param,
    invalid,
    get_timestep_weight,
)

logger = logging.getLogger(__name__)

ORDER = 20  # order in which to invoke constraints relative to other constraint files


def return_noconstraint(*args):
    logger.debug("group constraint returned NoConstraint: {}".format(",".join(args)))
    return po.Constraint.NoConstraint


def load_constraints(backend_model):
    model_data_dict = backend_model.__calliope_model_data["data"]

    for sense in ["min", "max", "equals"]:
        if "group_energy_cap_share_{}".format(sense) in model_data_dict:
            setattr(
                backend_model,
                "group_energy_cap_share_{}_constraint".format(sense),
                po.Constraint(
                    getattr(
                        backend_model, "group_names_energy_cap_share_{}".format(sense)
                    ),
                    [sense],
                    rule=energy_cap_share_constraint_rule,
                ),
            )

        if "group_energy_cap_{}".format(sense) in model_data_dict:
            setattr(
                backend_model,
                "group_energy_cap_{}_constraint".format(sense),
                po.Constraint(
                    getattr(backend_model, "group_names_energy_cap_{}".format(sense)),
                    [sense],
                    rule=energy_cap_constraint_rule,
                ),
            )

        if "group_resource_area_{}".format(sense) in model_data_dict:
            setattr(
                backend_model,
                "group_resource_area_{}_constraint".format(sense),
                po.Constraint(
                    getattr(
                        backend_model, "group_names_resource_area_{}".format(sense)
                    ),
                    [sense],
                    rule=resource_area_constraint_rule,
                ),
            )

        if "group_carrier_prod_{}".format(sense) in model_data_dict:
            setattr(
                backend_model,
                "group_carrier_prod_{}_constraint".format(sense),
                po.Constraint(
                    getattr(backend_model, "group_names_carrier_prod_{}".format(sense)),
                    [sense],
                    rule=carrier_prod_constraint_rule,
                ),
            )

        if "group_carrier_con_{}".format(sense) in model_data_dict:
            setattr(
                backend_model,
                "group_carrier_con_{}_constraint".format(sense),
                po.Constraint(
                    getattr(backend_model, "group_names_carrier_con_{}".format(sense)),
                    [sense],
                    rule=carrier_con_constraint_rule,
                ),
            )

        if "group_demand_share_{}".format(sense) in model_data_dict:
            setattr(
                backend_model,
                "group_demand_share_{}_constraint".format(sense),
                po.Constraint(
                    getattr(backend_model, "group_names_demand_share_{}".format(sense)),
                    [sense],
                    rule=demand_share_constraint_rule,
                ),
            )

        if "group_demand_share_per_timestep_{}".format(sense) in model_data_dict:
            setattr(
                backend_model,
                "group_demand_share_per_timestep_{}_constraint".format(sense),
                po.Constraint(
                    getattr(
                        backend_model,
                        "group_names_demand_share_per_timestep_{}".format(sense),
                    ),
                    backend_model.timesteps,
                    [sense],
                    rule=demand_share_per_timestep_constraint_rule,
                ),
            )

        if "group_carrier_prod_share_{}".format(sense) in model_data_dict:
            setattr(
                backend_model,
                "group_carrier_prod_share_{}_constraint".format(sense),
                po.Constraint(
                    getattr(
                        backend_model, "group_names_carrier_prod_share_{}".format(sense)
                    ),
                    [sense],
                    rule=carrier_prod_share_constraint_rule,
                ),
            )

        if "group_carrier_prod_share_per_timestep_{}".format(sense) in model_data_dict:
            setattr(
                backend_model,
                "group_carrier_prod_share_per_timestep_{}_constraint".format(sense),
                po.Constraint(
                    getattr(
                        backend_model,
                        "group_names_carrier_prod_share_per_timestep_{}".format(sense),
                    ),
                    backend_model.timesteps,
                    [sense],
                    rule=carrier_prod_share_per_timestep_constraint_rule,
                ),
            )

        if "group_net_import_share_{}".format(sense) in model_data_dict:
            setattr(
                backend_model,
                "group_net_import_share_{}_constraint".format(sense),
                po.Constraint(
                    getattr(
                        backend_model, "group_names_net_import_share_{}".format(sense)
                    ),
                    [sense],
                    rule=net_import_share_constraint_rule,
                ),
            )

        if "group_cost_{}".format(sense) in model_data_dict:
            setattr(
                backend_model,
                "group_cost_{}_constraint".format(sense),
                po.Constraint(
                    getattr(backend_model, "group_names_cost_{}".format(sense)),
                    backend_model.costs,
                    [sense],
                    rule=cost_cap_constraint_rule,
                ),
            )
        if "group_cost_var_{}".format(sense) in model_data_dict:
            setattr(
                backend_model,
                "group_cost_var_{}_constraint".format(sense),
                po.Constraint(
                    getattr(backend_model, "group_names_cost_var_{}".format(sense)),
                    backend_model.costs,
                    [sense],
                    rule=cost_var_cap_constraint_rule,
                ),
            )
        if "group_cost_investment_{}".format(sense) in model_data_dict:
            setattr(
                backend_model,
                "group_cost_investment_{}_constraint".format(sense),
                po.Constraint(
                    getattr(
                        backend_model, "group_names_cost_investment_{}".format(sense)
                    ),
                    backend_model.costs,
                    [sense],
                    rule=cost_investment_cap_constraint_rule,
                ),
            )

    if "group_demand_share_per_timestep_decision" in model_data_dict:
        backend_model.group_demand_share_per_timestep_decision_main_constraint = (
            po.Constraint(
                backend_model.group_names_demand_share_per_timestep_decision,
                backend_model.loc_tech_carriers_demand_share_per_timestep_lhs,
                backend_model.timesteps,
                rule=demand_share_per_timestep_decision_main_constraint_rule,
            )
        )
        backend_model.group_demand_share_per_timestep_decision_sum_constraint = (
            po.Constraint(
                backend_model.group_names_demand_share_per_timestep_decision,
                rule=demand_share_per_timestep_decision_sum_constraint_rule,
            )
        )


def equalizer(lhs, rhs, sign):
    if sign == "max":
        return lhs <= rhs
    elif sign == "min":
        return lhs >= rhs
    elif sign == "equals":
        return lhs == rhs
    else:
        raise ValueError("Invalid sign: {}".format(sign))


def get_group_lhs_loc_tech_carriers(backend_model, group_name):
    return getattr(
        backend_model, "group_constraint_loc_tech_carriers_{}".format(group_name)
    )


def get_group_lhs_and_rhs_loc_tech_carriers(backend_model, group_name):
    lhs_loc_tech_carriers = getattr(
        backend_model, "group_constraint_loc_tech_carriers_{}_lhs".format(group_name)
    )
    rhs_loc_tech_carriers = getattr(
        backend_model, "group_constraint_loc_tech_carriers_{}_rhs".format(group_name)
    )
    return lhs_loc_tech_carriers, rhs_loc_tech_carriers


def demand_share_constraint_rule(backend_model, group_name, what):
    """
    Enforces shares of demand of a carrier to be met by the given groups
    of technologies at the given locations, on average over the entire
    model period. The share is relative to ``demand`` technologies only.

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{loc::tech::carrier \\in given\\_group, timestep \\in timesteps} carrier_{prod}(loc::tech::carrier, timestep) \\leq
            share \\times \\sum_{loc::tech:carrier \\in loc\\_techs\\_demand \\in given\\_locations, timestep\\in timesteps}
            carrier_{con}(loc::tech::carrier, timestep)

    """
    share = get_param(backend_model, "group_demand_share_{}".format(what), (group_name))

    if invalid(share):
        return return_noconstraint("demand_share", group_name)
    else:
        (
            lhs_loc_tech_carriers,
            rhs_loc_tech_carriers,
        ) = get_group_lhs_and_rhs_loc_tech_carriers(backend_model, group_name)

        lhs = sum(
            backend_model.carrier_prod[loc_tech_carrier, timestep]
            for loc_tech_carrier in lhs_loc_tech_carriers
            for timestep in backend_model.timesteps
        )

        rhs = (
            share
            * -1
            * sum(
                backend_model.carrier_con[loc_tech_carrier, timestep]
                for loc_tech_carrier in rhs_loc_tech_carriers
                for timestep in backend_model.timesteps
            )
        )

        return equalizer(lhs, rhs, what)


def demand_share_per_timestep_constraint_rule(
    backend_model, group_name, timestep, what
):
    """
    Enforces shares of demand of a carrier to be met by the given groups
    of technologies at the given locations, in each timestep.
    The share is relative to ``demand`` technologies only.

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{loc::tech::carrier \\in given\\_group} carrier_{prod}(loc::tech::carrier, timestep) \\leq
            share \\times \\sum_{loc::tech:carrier \\in loc\\_techs\\_demand \\in given\\_locations}
            carrier_{con}(loc::tech::carrier, timestep) for timestep \\in timesteps

    """
    share = get_param(
        backend_model,
        "group_demand_share_per_timestep_{}".format(what),
        (group_name),
    )

    if invalid(share):
        return return_noconstraint("demand_share_per_timestep", group_name)
    else:
        (
            lhs_loc_tech_carriers,
            rhs_loc_tech_carriers,
        ) = get_group_lhs_and_rhs_loc_tech_carriers(backend_model, group_name)

        lhs = sum(
            backend_model.carrier_prod[loc_tech_carrier, timestep]
            for loc_tech_carrier in lhs_loc_tech_carriers
        )

        rhs = (
            share
            * -1
            * sum(
                backend_model.carrier_con[loc_tech_carrier, timestep]
                for loc_tech_carrier in rhs_loc_tech_carriers
            )
        )

        return equalizer(lhs, rhs, what)


def demand_share_per_timestep_decision_main_constraint_rule(
    backend_model, group_name, loc_tech_carrier, timestep
):
    """
    Allows the model to decide on how a fraction demand for a carrier is met
    by the given groups, which will all have the same share in each timestep.
    The share is relative to the actual demand from ``demand`` technologies only.

    The main constraint enforces that the shares are the same in each timestep.

    .. container:: scrolling-wrapper

        .. math::
            \\sum_{loc::tech::carrier \\in given\\_group} carrier_{prod}(loc::tech::carrier, timestep)

            =

            \\sum_{loc::tech::carrier \\in given\\_group}
            required\\_resource(loc::tech::carrier, timestep)

            \\times \\sum_{loc::tech::carrier \\in given\\_group}
            demand\\_share\\_per\\_timestep\\_decision(loc::tech::carrier)

            \\forall timestep \\in timesteps

            \\forall tech \\in techs

    """
    (
        _,
        rhs_loc_tech_carriers,
    ) = get_group_lhs_and_rhs_loc_tech_carriers(backend_model, group_name)
    lhs = backend_model.carrier_prod[loc_tech_carrier, timestep]
    rhs = (
        -1
        * sum(
            backend_model.required_resource[
                rhs_loc_tech_carrier.rsplit("::", 1)[0], timestep
            ]
            for rhs_loc_tech_carrier in rhs_loc_tech_carriers
        )
        * backend_model.demand_share_per_timestep_decision[loc_tech_carrier]
    )

    return equalizer(lhs, rhs, "min")


def demand_share_per_timestep_decision_sum_constraint_rule(backend_model, group_name):
    """
       Allows the model to decide on how a fraction of demand for a carrier is met
       by the given groups, which will all have the same share in each timestep.
       The share is relative to the actual demand from ``demand`` technologies only.

       The sum constraint ensures that all decision shares add up to the share of
       carrier demand specified in the constraint.

       This constraint is only applied if the share of carrier demand has been
       set to a not-None value.

    .. container:: scrolling-wrapper

           .. math::
               share = \\sum_{loc::tech::carrier \\in given\\_group}
               demand\\_share\\_per\\_timestep\\_decision(loc::tech::carrier)


    """
    share_of_carrier_demand = get_param(
        backend_model, "group_demand_share_per_timestep_decision", (group_name)
    )

    # If inf was given that means that we don't limit the total share
    if invalid(share_of_carrier_demand) or np.isinf(po.value(share_of_carrier_demand)):
        return return_noconstraint("demand_share_per_timestep_decision_sum", group_name)
    else:
        lhs_loc_tech_carriers, _ = get_group_lhs_and_rhs_loc_tech_carriers(
            backend_model, group_name
        )

        return share_of_carrier_demand == sum(
            backend_model.demand_share_per_timestep_decision[loc_tech_carrier]
            for loc_tech_carrier in lhs_loc_tech_carriers
        )


def carrier_prod_share_constraint_rule(backend_model, group_name, what):
    """
    Enforces shares of carrier_prod for groups of technologies and locations,
    on average over the entire model period. The share is relative to
    ``supply`` and ``supply_plus`` technologies only.

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{loc::tech::carrier \\in given\\_group, timestep \\in timesteps} carrier_{prod}(loc::tech::carrier, timestep) \\leq
            share \\times \\sum_{loc::tech:carrier \\in loc\\_tech\\_carriers\\_supply\\_all \\in given\\_locations, timestep\\in timesteps}
            carrier_{prod}(loc::tech::carrier, timestep)

    """
    share = get_param(
        backend_model, "group_carrier_prod_share_{}".format(what), (group_name)
    )

    if invalid(share):
        return return_noconstraint("carrier_prod_share", group_name)
    else:
        (
            lhs_loc_tech_carriers,
            rhs_loc_tech_carriers,
        ) = get_group_lhs_and_rhs_loc_tech_carriers(backend_model, group_name)

        lhs = sum(
            backend_model.carrier_prod[loc_tech_carrier, timestep]
            for loc_tech_carrier in lhs_loc_tech_carriers
            for timestep in backend_model.timesteps
        )
        rhs = share * sum(
            backend_model.carrier_prod[loc_tech_carrier, timestep]
            for loc_tech_carrier in rhs_loc_tech_carriers
            for timestep in backend_model.timesteps
        )

        return equalizer(lhs, rhs, what)


def carrier_prod_share_per_timestep_constraint_rule(
    backend_model, group_name, timestep, what
):
    """
    Enforces shares of carrier_prod for groups of technologies and locations,
    in each timestep. The share is relative to ``supply`` and ``supply_plus``
    technologies only.

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{loc::tech::carrier \\in given\\_group} carrier_{prod}(loc::tech::carrier, timestep) \\leq
            share \\times \\sum_{loc::tech:carrier \\in loc\\_tech\\_carriers\\_supply\\_all \\in given\\_locations}
            carrier_{prod}(loc::tech::carrier, timestep) for timestep \\in timesteps

    """
    share = get_param(
        backend_model,
        "group_carrier_prod_share_per_timestep_{}".format(what),
        (group_name),
    )

    if invalid(share):
        return return_noconstraint("carrier_prod_share_per_timestep", group_name)
    else:
        (
            lhs_loc_tech_carriers,
            rhs_loc_tech_carriers,
        ) = get_group_lhs_and_rhs_loc_tech_carriers(backend_model, group_name)

        lhs = sum(
            backend_model.carrier_prod[loc_tech_carrier, timestep]
            for loc_tech_carrier in lhs_loc_tech_carriers
        )
        rhs = share * sum(
            backend_model.carrier_prod[loc_tech_carrier, timestep]
            for loc_tech_carrier in rhs_loc_tech_carriers
        )

        return equalizer(lhs, rhs, what)


def net_import_share_constraint_rule(backend_model, group_name, what):
    """
    Enforces demand shares of net imports from transmission technologies for groups of locations,
    on average over the entire model period. Transmission within the group are ignored. The share
    is relative to ``demand`` technologies only.

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{loc::tech::carrier \\in loc\\_tech\\_carriers\\_transmission \\in given\\_locations, timestep \\in timesteps} carrier_{prod}(loc::tech::carrier, timestep)
            + \\sum_{loc::tech::carrier \\in loc\\_tech\\_carriers\\_transmission \\in given\\_locations, timestep \\in timesteps} carrier_{con}(loc::tech::carrier, timestep) \\leq
            share \\times \\sum_{loc::tech:carrier \\in loc\\_tech\\_demand \\in given\\_locations, timestep\\in timesteps}
            carrier_{con}(loc::tech::carrier, timestep)

    """
    share = get_param(backend_model, f"group_net_import_share_{what}", (group_name))

    if invalid(share):
        return return_noconstraint("net_import_share", group_name)
    else:
        (
            lhs_loc_tech_carriers,
            rhs_loc_tech_carriers,
        ) = get_group_lhs_and_rhs_loc_tech_carriers(backend_model, group_name)

        lhs = sum(
            (
                backend_model.carrier_prod[loc_tech_carrier, timestep]
                + backend_model.carrier_con[loc_tech_carrier, timestep]
            )
            for loc_tech_carrier in lhs_loc_tech_carriers
            for timestep in backend_model.timesteps
        )
        rhs = -share * sum(
            backend_model.carrier_con[loc_tech_carrier, timestep]
            for loc_tech_carrier in rhs_loc_tech_carriers
            for timestep in backend_model.timesteps
        )
        return equalizer(lhs, rhs, what)


def carrier_prod_constraint_rule(backend_model, group_name, what):
    """
    Enforces carrier_prod for groups of technologies and locations,
    as a sum over the entire model period.

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{loc::tech::carrier \\in given\\_group, timestep \\in timesteps} carrier_{prod}(loc::tech::carrier, timestep) \\leq carrier_prod_max

    """
    limit = get_param(backend_model, f"group_carrier_prod_{what}", (group_name))

    if invalid(limit):
        return return_noconstraint("carrier_prod", group_name)
    else:
        lhs_loc_tech_carriers = get_group_lhs_loc_tech_carriers(
            backend_model, group_name
        )

        lhs = sum(
            backend_model.carrier_prod[loc_tech_carrier, timestep]
            for loc_tech_carrier in lhs_loc_tech_carriers
            for timestep in backend_model.timesteps
        )
        return equalizer(lhs, limit, what)


def carrier_con_constraint_rule(backend_model, constraint_group, what):
    """
    Enforces carrier_con for groups of technologies and locations,
    as a sum over the entire model period. limits are always negative, so min/max
    is relative to zero (i.e. min = -1 means carrier_con must be -1 or less)

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{loc::tech::carrier \\in given\\_group, timestep \\in timesteps} carrier_{con}(loc::tech::carrier, timestep) \\geq carrier_con_max

    """
    limit = get_param(
        backend_model, "group_carrier_con_{}".format(what), (constraint_group)
    )

    if invalid(limit):
        return return_noconstraint("carrier_con", constraint_group)
    else:
        lhs_loc_tech_carriers = get_group_lhs_loc_tech_carriers(
            backend_model, constraint_group
        )

        lhs = sum(
            backend_model.carrier_con[loc_tech_carrier, timestep]
            for loc_tech_carrier in lhs_loc_tech_carriers
            for timestep in backend_model.timesteps
        )

        return equalizer(limit, lhs, what)


def energy_cap_share_constraint_rule(backend_model, constraint_group, what):
    """
    Enforces shares of energy_cap for groups of technologies and locations. The
    share is relative to ``supply``, ``supply_plus``, ``conversion``, and ``conversion_plus``
    technologies only.

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{loc::tech \\in given\\_group} energy_{cap}(loc::tech) \\leq
            share \\times \\sum_{loc::tech \\in loc\\_tech\\_supply\\_all \\in given\\_locations} energy_{cap}(loc::tech)
    """
    share = get_param(
        backend_model, "group_energy_cap_share_{}".format(what), (constraint_group)
    )

    if invalid(share):
        return return_noconstraint("energy_cap_share", constraint_group)
    else:
        lhs_loc_techs = getattr(
            backend_model, "group_constraint_loc_techs_{}".format(constraint_group)
        )
        lhs_locs = [loc_tech.split("::")[0] for loc_tech in lhs_loc_techs]
        rhs_loc_techs = [
            i
            for i in backend_model.loc_techs_supply_conversion_all
            if i.split("::")[0] in lhs_locs
        ]

        lhs = sum(backend_model.energy_cap[loc_tech] for loc_tech in lhs_loc_techs)
        rhs = share * sum(
            backend_model.energy_cap[loc_tech] for loc_tech in rhs_loc_techs
        )

        return equalizer(lhs, rhs, what)


def energy_cap_constraint_rule(backend_model, constraint_group, what):
    """
    Enforce upper and lower bounds for energy_cap of energy_cap
    for groups of technologies and locations.

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{loc::tech \\in given\\_group} energy_{cap}(loc::tech) \\leq energy\\_cap\\_max\\\\

            \\sum_{loc::tech \\in given\\_group} energy_{cap}(loc::tech) \\geq energy\\_cap\\_min

    """
    threshold = get_param(
        backend_model, "group_energy_cap_{}".format(what), (constraint_group)
    )

    if invalid(threshold):
        return return_noconstraint("energy_cap", constraint_group)
    else:
        lhs_loc_techs = getattr(
            backend_model, "group_constraint_loc_techs_{}".format(constraint_group)
        )

        # Transmission techs only contribute half their capacity in each direction
        lhs = []
        for loc_tech in lhs_loc_techs:
            if loc_tech_is_in(backend_model, loc_tech, "loc_techs_transmission"):
                weight = 0.5
            else:
                weight = 1

            lhs.append(weight * backend_model.energy_cap[loc_tech])

        rhs = threshold

        return equalizer(sum(lhs), rhs, what)


def storage_cap_constraint_rule(backend_model, constraint_group, what):
    """
    Enforce upper and lower bounds for storage_cap of storage_cap
    for groups of technologies and locations.

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{loc::tech \\in given\\_group} storage_{cap}(loc::tech) \\leq storage\\_cap\\_max\\\\

            \\sum_{loc::tech \\in given\\_group} storage_{cap}(loc::tech) \\geq storage\\_cap\\_min

    """
    threshold = get_param(
        backend_model, "group_storage_cap_{}".format(what), (constraint_group)
    )

    if invalid(threshold):
        return return_noconstraint("storage_cap", constraint_group)
    else:
        lhs_loc_techs = getattr(
            backend_model, "group_constraint_loc_techs_{}".format(constraint_group)
        )

        lhs = sum(backend_model.storage_cap[loc_tech] for loc_tech in lhs_loc_techs)
        rhs = threshold

        return equalizer(lhs, rhs, what)


def cost_cap_constraint_rule(backend_model, group_name, cost, what):
    """
    Limit cost for a specific cost class to a certain value,
    i.e. Ɛ-constrained costs,
    for groups of technologies and locations.

    .. container:: scrolling-wrapper

        .. math::

            \\sum{loc::tech \\in loc\\_techs_{group\\_name}, timestep \\in timesteps}
            \\boldsymbol{cost}(cost, loc::tech, timestep)
            \\begin{cases}
                \\leq cost\\_max(cost)
                \\geq cost\\_min(cost)
                = cost\\_equals(cost)
            \\end{cases}

    """
    cost_cap = get_param(
        backend_model, "group_cost_{}".format(what), (cost, group_name)
    )
    if invalid(cost_cap):
        return return_noconstraint("cost_cap", group_name)
    else:
        loc_techs = [
            i
            for i in getattr(
                backend_model, "group_constraint_loc_techs_{}".format(group_name)
            )
            if i in backend_model.loc_techs_cost
        ]

        sum_cost = sum(backend_model.cost[cost, loc_tech] for loc_tech in loc_techs)

        return equalizer(sum_cost, cost_cap, what)


def cost_investment_cap_constraint_rule(backend_model, group_name, cost, what):
    """
    Limit investment costs specific to a cost class to a
    certain value, i.e. Ɛ-constrained costs,
    for groups of technologies and locations.

    .. container:: scrolling-wrapper

        .. math::

            \\sum{loc::tech \\in loc\\_techs_{group\\_name}, timestep \\in timesteps}
            \\boldsymbol{cost\\_{investment}}(cost, loc::tech, timestep)
            \\begin{cases}
                \\leq cost\\_investment\\_max(cost)
                \\geq cost\\_investment\\_min(cost)
                = cost\\_investment\\_equals(cost)
            \\end{cases}

    """
    cost_cap = get_param(
        backend_model, "group_cost_investment_{}".format(what), (cost, group_name)
    )

    if invalid(cost_cap):
        return return_noconstraint("cost_investment_cap", group_name)
    else:
        loc_techs = [
            i
            for i in getattr(
                backend_model, "group_constraint_loc_techs_{}".format(group_name)
            )
            if i in backend_model.loc_techs_investment_cost
        ]

        sum_cost = sum(
            backend_model.cost_investment[cost, loc_tech] for loc_tech in loc_techs
        )

        return equalizer(sum_cost, cost_cap, what)


def cost_var_cap_constraint_rule(backend_model, group_name, cost, what):
    """
    Limit variable costs specific to a cost class
    to a certain value, i.e. Ɛ-constrained costs,
    for groups of technologies and locations.

    .. container:: scrolling-wrapper

        .. math::

            \\sum{loc::tech \\in loc\\_techs_{group\\_name}, timestep \\in timesteps}
            \\boldsymbol{cost\\_{var}}(cost, loc::tech, timestep)
            \\begin{cases}
                \\leq cost\\_var\\_max(cost)
                \\geq cost\\_var\\_min(cost)
                = cost\\_var\\_equals(cost)
            \\end{cases}

    """
    cost_cap = get_param(
        backend_model, "group_cost_var_{}".format(what), (cost, group_name)
    )

    if invalid(cost_cap):
        return return_noconstraint("cost_var_cap", group_name)
    else:
        loc_techs = [
            i
            for i in getattr(
                backend_model, "group_constraint_loc_techs_{}".format(group_name)
            )
            if i in backend_model.loc_techs_om_cost
        ]

        sum_cost = sum(
            backend_model.cost_var[cost, loc_tech, timestep]
            for loc_tech in loc_techs
            for timestep in backend_model.timesteps
        )

        return equalizer(sum_cost, cost_cap, what)


def resource_area_constraint_rule(backend_model, constraint_group, what):
    """
    Enforce upper and lower bounds of resource_area for groups of
    technologies and locations.

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{resource_{area}}(loc::tech) \\leq group\\_resource\\_area\\_max\\\\

            \\boldsymbol{resource_{area}}(loc::tech) \\geq group\\_resource\\_area\\_min

    """
    threshold = get_param(
        backend_model, "group_resource_area_{}".format(what), (constraint_group)
    )

    if invalid(threshold):
        return return_noconstraint("resource_area", constraint_group)
    else:
        lhs_loc_techs = getattr(
            backend_model, "group_constraint_loc_techs_{}".format(constraint_group)
        )

        lhs = sum(backend_model.resource_area[loc_tech] for loc_tech in lhs_loc_techs)
        rhs = threshold

        return equalizer(lhs, rhs, what)



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.

Licensed under the Apache 2.0 License (see LICENSE file).

objective.py
~~~~~~~~~~~~

Objective functions.

"""

import pyomo.core as po  # pylint: disable=import-error
from calliope.core.util.tools import load_function


def minmax_cost_optimization(backend_model):
    """
    Minimize or maximise total system cost for specified cost class or a set of cost classes.
    cost_class is a string or dictionary. If a string, it is automatically converted to a
    dictionary with a single key:value pair where value == 1. The dictionary provides a weight
    for each cost class of interest: {cost_1: weight_1, cost_2: weight_2, etc.}.

    If unmet_demand is in use, then the calculated cost of unmet_demand is
    added or subtracted from the total cost in the opposite sense to the
    objective.

    .. container:: scrolling-wrapper

        .. math::

            min: z = \\sum_{loc::tech_{cost},k} (cost(loc::tech, cost=cost_{k}) \\times weight_{k}) +
             \\sum_{loc::carrier,timestep} (unmet\\_demand(loc::carrier, timestep) \\times bigM)

            max: z = \\sum_{loc::tech_{cost},k} (cost(loc::tech, cost=cost_{k}) \\times weight_{k}) -
             \\sum_{loc::carrier,timestep} (unmet\\_demand(loc::carrier, timestep) \\times bigM)

    """

    def obj_rule(backend_model):
        if backend_model.__calliope_run_config.get("ensure_feasibility", False):
            unmet_demand = (
                sum(
                    (
                        backend_model.unmet_demand[loc_carrier, timestep]
                        - backend_model.unused_supply[loc_carrier, timestep]
                    )
                    * backend_model.timestep_weights[timestep]
                    for loc_carrier in backend_model.loc_carriers
                    for timestep in backend_model.timesteps
                )
                * backend_model.bigM
            )
            if backend_model.objective_sense == "maximize":
                unmet_demand *= -1
        else:
            unmet_demand = 0

        return (
            sum(
                backend_model.cost[k, loc_tech] * v
                for loc_tech in backend_model.loc_techs_cost
                for k, v in backend_model.objective_cost_class.items()
            )
            + unmet_demand
        )

    backend_model.obj = po.Objective(
        sense=load_function("pyomo.core." + backend_model.objective_sense),
        rule=obj_rule,
    )
    backend_model.obj.domain = po.Reals


def check_feasibility(backend_model):
    """
    Dummy objective, to check that there are no conflicting constraints.

    .. container:: scrolling-wrapper

        .. math::

            min: z = 1

    """

    def obj_rule(backend_model):
        return 1

    backend_model.obj = po.Objective(sense=po.minimize, rule=obj_rule)
    backend_model.obj.domain = po.Reals



import os
import tempfile

import pytest  # pylint: disable=unused-import
from click.testing import CliRunner

import calliope
from calliope import cli, AttrDict

_THIS_DIR = os.path.dirname(__file__)
_MODEL_NATIONAL = os.path.join(
    _THIS_DIR, "..", "example_models", "national_scale", "model.yaml"
)
_MINIMAL_TEST_MODEL = os.path.join(
    _THIS_DIR, "common", "test_model", "model_minimal.yaml"
)


class TestCLI:
    def test_new(self):
        runner = CliRunner()
        with tempfile.TemporaryDirectory() as tempdir:
            new_path = os.path.join(tempdir, "test")
            result = runner.invoke(cli.new, [new_path])
            assert result.exit_code == 0
            # Assert that `model.yaml` in the target dir exists
            assert os.path.isfile(os.path.join(tempdir, "test", "model.yaml"))

    def test_run_from_yaml(self):
        runner = CliRunner()

        with runner.isolated_filesystem() as tempdir:
            result = runner.invoke(
                cli.run,
                [
                    _MODEL_NATIONAL,
                    "--save_netcdf=output.nc",
                    "--save_plots=results.html",
                ],
            )
            assert result.exit_code == 0
            assert os.path.isfile(os.path.join(tempdir, "output.nc"))
            assert os.path.isfile(os.path.join(tempdir, "results.html"))

    def test_save_per_spore(self):
        runner = CliRunner()

        with runner.isolated_filesystem() as tempdir:
            os.mkdir(os.path.join(tempdir, "output"))
            print(os.listdir(tempdir))
            result = runner.invoke(
                cli.run,
                [
                    _MODEL_NATIONAL,
                    "--save_netcdf=output.nc",
                    "--scenario=spores",
                    "--override_dict={'run.spores_options.save_per_spore': True}",
                ],
            )
            print(os.listdir(os.path.join(tempdir, "output")))
            assert result.exit_code == 0
            for i in ["0", "1", "2", "3"]:
                assert os.path.isfile(os.path.join(tempdir, "output", f"spore_{i}.nc"))
            assert os.path.isfile(os.path.join(tempdir, "output.nc"))

    def test_incorrect_file_format(self):
        runner = CliRunner()

        result = runner.invoke(cli.run, ["test_model.txt", "--save_netcdf=output.nc"])
        assert "Cannot determine model file format" in result.output
        assert result.exit_code == 1

    def test_incorrect_model_format(self):
        runner = CliRunner()

        result = runner.invoke(
            cli.run, ["test_model.txt", "--model_format=yml", "--save_netcdf=output.nc"]
        )
        assert "Invalid model format" in result.output
        assert result.exit_code == 1

    @pytest.mark.parametrize(
        "arg", (("--scenario=test"), ("--override_dict={'model.name': 'test'}"))
    )
    def test_unavailable_arguments(self, arg):
        runner = CliRunner()

        result = runner.invoke(
            cli.run, ["test_model.nc", arg, "--save_netcdf=output.nc"]
        )
        assert (
            "the --scenario and --override_dict options are not available"
            in result.output
        )
        assert result.exit_code == 1

    def test_run_from_netcdf(self):
        runner = CliRunner()
        model = calliope.examples.national_scale()
        with runner.isolated_filesystem() as tempdir:
            model_file = os.path.join(tempdir, "model.nc")
            out_file = os.path.join(tempdir, "output.nc")
            model.to_netcdf(model_file)
            result = runner.invoke(cli.run, [model_file, "--save_netcdf=output.nc"])
            assert result.exit_code == 0
            assert os.path.isfile(out_file)

    def test_run_save_lp(self):
        runner = CliRunner()

        with runner.isolated_filesystem() as tempdir:
            result = runner.invoke(cli.run, [_MODEL_NATIONAL, "--save_lp=output.lp"])
            assert result.exit_code == 0
            assert os.path.isfile(os.path.join(tempdir, "output.lp"))

    def test_generate_runs_bash(self):
        runner = CliRunner()

        with runner.isolated_filesystem() as tempdir:
            result = runner.invoke(
                cli.generate_runs,
                [
                    _MODEL_NATIONAL,
                    "test.sh",
                    "--kind=bash",
                    '--scenarios="run1;run2;run3;run4"',
                ],
            )
            assert result.exit_code == 0
            assert os.path.isfile(os.path.join(tempdir, "test.sh"))

    def test_generate_runs_windows(self):
        runner = CliRunner()

        with runner.isolated_filesystem() as tempdir:
            result = runner.invoke(
                cli.generate_runs,
                [
                    _MODEL_NATIONAL,
                    "test.bat",
                    "--kind=windows",
                    '--scenarios="run1;run2;run3;run4"',
                ],
            )
            assert result.exit_code == 0
            assert os.path.isfile(os.path.join(tempdir, "test.bat"))

    def test_generate_runs_bsub(self):
        runner = CliRunner()

        with runner.isolated_filesystem() as tempdir:
            result = runner.invoke(
                cli.generate_runs,
                [
                    _MODEL_NATIONAL,
                    "test.sh",
                    "--kind=bsub",
                    '--scenarios="run1;run2;run3;run4"',
                    "--cluster_mem=1G",
                    "--cluster_time=100",
                ],
            )
            assert result.exit_code == 0
            assert os.path.isfile(os.path.join(tempdir, "test.sh"))
            assert os.path.isfile(os.path.join(tempdir, "test.sh.array.sh"))

    def test_generate_runs_sbatch(self):
        runner = CliRunner()

        with runner.isolated_filesystem() as tempdir:
            result = runner.invoke(
                cli.generate_runs,
                [
                    _MODEL_NATIONAL,
                    "test.sh",
                    "--kind=sbatch",
                    '--scenarios="run1;run2;run3;run4"',
                    "--cluster_mem=1G",
                    "--cluster_time=100",
                ],
            )
            assert result.exit_code == 0
            assert os.path.isfile(os.path.join(tempdir, "test.sh"))
            assert os.path.isfile(os.path.join(tempdir, "test.sh.array.sh"))

    def test_debug(self):
        runner = CliRunner()
        result = runner.invoke(cli.run, ["foo.yaml", "--debug"])
        assert result.exit_code == 1
        assert "Traceback (most recent call last)" in result.output

        result = runner.invoke(cli.run, ["foo.yaml"])
        assert result.exit_code == 1
        assert "Traceback (most recent call last)" not in result.output

    def test_generate_scenarios(self):
        runner = CliRunner()
        with runner.isolated_filesystem() as tempdir:
            out_file = os.path.join(tempdir, "scenarios.yaml")
            result = runner.invoke(
                cli.generate_scenarios,
                [
                    _MODEL_NATIONAL,
                    out_file,
                    "cold_fusion",
                    "run1;run2",
                    "cold_fusion_cap_share;cold_fusion_prod_share",
                ],
            )
            assert result.exit_code == 0
            assert os.path.isfile(out_file)
            scenarios = AttrDict.from_yaml(out_file)
            assert "scenario_0" not in scenarios["scenarios"]
            assert scenarios["scenarios"]["scenario_1"] == [
                "cold_fusion",
                "run1",
                "cold_fusion_cap_share",
            ]

    def test_no_success_exit_code_when_infeasible(self):
        runner = CliRunner()
        result = runner.invoke(
            cli.run,
            [
                _MINIMAL_TEST_MODEL,
                "--scenario=investment_costs",  # without these, the model cannot run
                "--override_dict={techs.test_supply_elec.constraints.energy_cap_max: 1}",  # elec supply too low
            ],
        )
        assert result.exit_code != 0

    def test_success_exit_code_when_infeasible_and_demanded(self):
        runner = CliRunner()
        result = runner.invoke(
            cli.run,
            [
                _MINIMAL_TEST_MODEL,
                "--no_fail_when_infeasible",
                "--scenario=investment_costs",  # without these, the model cannot run
                "--override_dict={techs.test_supply_elec.constraints.energy_cap_max: 1}",  # elec supply too low
            ],
        )
        assert result.exit_code == 0



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

postprocess.py
~~~~~~~~~~~~~~

Functionality to post-process model results.

"""

import logging

import xarray as xr
import numpy as np

from calliope.core.util.dataset import split_loc_techs
from calliope.core.util.logging import log_time
from calliope.core.attrdict import AttrDict
from calliope.preprocess.util import concat_iterable

logger = logging.getLogger(__name__)


def postprocess_model_results(results, model_data, timings):
    """
    Adds additional post-processed result variables to
    the given model results in-place. Model must have solved successfully.

    Parameters
    ----------
    results : xarray Dataset
        Output from the solver backend
    model_data : xarray Dataset
        Calliope model data, stored as calliope.Model()._model_data
    timings : dict
        Calliope timing dictionary, stored as calliope.Model()._timings

    Returns
    -------
    results : xarray Dataset
        Input results Dataset, with additional DataArray variables and removed
        all instances of unreasonably low numbers (set by zero_threshold)

    """
    log_time(logger, timings, "post_process_start", comment="Postprocessing: started")

    run_config = AttrDict.from_yaml_string(model_data.attrs["run_config"])
    results["capacity_factor"] = capacity_factor(results, model_data)
    results["systemwide_capacity_factor"] = capacity_factor(
        results, model_data, systemwide=True
    )
    results["systemwide_levelised_cost"] = systemwide_levelised_cost(
        results, model_data
    )
    results["total_levelised_cost"] = systemwide_levelised_cost(
        results, model_data, total=True
    )
    results = clean_results(results, run_config.get("zero_threshold", 0), timings)

    for var_data in results.data_vars.values():
        if "is_result" not in var_data.attrs.keys():
            var_data.attrs["is_result"] = 1

    log_time(
        logger,
        timings,
        "post_process_end",
        time_since_run_start=True,
        comment="Postprocessing: ended",
    )

    if "run_solution_returned" in timings.keys():
        results.attrs["solution_time"] = (
            timings["run_solution_returned"] - timings["run_start"]
        ).total_seconds()
        results.attrs["time_finished"] = timings["run_solution_returned"].strftime(
            "%Y-%m-%d %H:%M:%S"
        )

    return results


def capacity_factor(results, model_data, systemwide=False):
    """
    Returns a DataArray with capacity factor for the given results.
    The results are either indexed by loc_tech_carriers_prod and timesteps,
    or by techs and carriers if systemwide results are being calculated.

    The weight of timesteps is considered when computing systemwide capacity factors,
    such that higher-weighted timesteps have a stronger influence
    on the resulting system-wide time-averaged capacity factor.

    """
    # In operate mode, energy_cap is an input parameter
    if "energy_cap" not in results.keys():
        energy_cap = model_data.energy_cap
    else:
        energy_cap = results.energy_cap

    _prod = split_loc_techs(results["carrier_prod"])
    _cap = split_loc_techs(energy_cap)
    if systemwide:
        # Aggregated/clustered days are represented `timestep_weights` times
        prod_sum = (_prod * model_data.timestep_weights).sum(["timesteps", "locs"])
        cap_sum = _cap.sum(dim="locs")
        time_sum = (model_data.timestep_resolution * model_data.timestep_weights).sum()
        capacity_factors = prod_sum / (cap_sum * time_sum)

    else:
        extra_dims = {
            i: model_data[i].to_index() for i in _prod.dims if i not in _cap.dims
        }
        capacity_factors = (
            (_prod / _cap.expand_dims(extra_dims))
            .fillna(0)
            .stack({"loc_tech_carriers_prod": ["locs", "techs", "carriers"]})
        )
        new_idx = concat_iterable(
            capacity_factors.loc_tech_carriers_prod.values, ["::", "::"]
        )
        capacity_factors = capacity_factors.assign_coords(
            {"loc_tech_carriers_prod": new_idx}
        ).reindex({"loc_tech_carriers_prod": results.loc_tech_carriers_prod})

    return capacity_factors


def systemwide_levelised_cost(results, model_data, total=False):
    """
    Returns a DataArray with systemwide levelised costs for the given
    results, indexed by techs, carriers and costs if total is False,
    or by carriers and costs if total is True.

    The weight of timesteps is considered when computing levelised costs:

    * costs are already multiplied by weight in the constraints, and not
      further adjusted here.

    * production is not multiplied by weight in the contraints, so scaled
      by weight here to be consistent with costs. CAUTION: this scaling
      is temporary duriing levelised cost computation - the actual
      costs in the results remain untouched.

    Parameters
    ----------
    results : xarray.Dataset
        Model results
    model_data : xarray.Dataset
        Model input data
    total : bool, optional
        If False (default) returns per-technology levelised cost, if True,
        returns overall system-wide levelised cost.

    """
    cost = results["cost"]
    # Here we scale production by timestep weight
    carrier_prod = results["carrier_prod"] * model_data.timestep_weights

    if total:
        cost = split_loc_techs(cost).sum(dim=["locs", "techs"])
        supply_only_carrier_prod = carrier_prod.sel(
            loc_tech_carriers_prod=list(
                model_data.loc_tech_carriers_supply_conversion_all.values
            )
        )
        carrier_prod = split_loc_techs(supply_only_carrier_prod).sum(
            dim=["timesteps", "locs", "techs"]
        )
    else:
        cost = split_loc_techs(cost).sum(dim=["locs"])
        carrier_prod = split_loc_techs(carrier_prod).sum(["timesteps", "locs"])

    levelised_cost = []

    for carrier in carrier_prod["carriers"].values:
        levelised_cost.append(cost / carrier_prod.loc[dict(carriers=carrier)])

    return xr.concat(levelised_cost, dim="carriers")


def clean_results(results, zero_threshold, timings):
    """
    Remove unreasonably small values (solver output can lead to floating point
    errors) and remove unmet_demand if it was never used (i.e. sum = zero)

    zero_threshold is a value set in model configuration. If not set, defaults
    to zero (i.e. doesn't do anything). Reasonable value = 1e-12
    """
    threshold_applied = []
    for k, v in results.data_vars.items():
        # If there are any values in the data variable which fall below the
        # threshold, note the data variable name and set those values to zero
        if v.where(abs(v) < zero_threshold, drop=True).sum():
            threshold_applied.append(k)
            with np.errstate(invalid="ignore"):
                v.values[abs(v.values) < zero_threshold] = 0
            v.loc[{}] = v.values

    if threshold_applied:
        comment = "All values < {} set to 0 in {}".format(
            zero_threshold, ", ".join(threshold_applied)
        )
    else:
        comment = "zero threshold of {} not required".format(zero_threshold)

    log_time(logger, timings, "threshold_applied", comment="Postprocessing: " + comment)

    # Combine unused_supply and unmet_demand into one variable
    if (
        "unmet_demand" in results.data_vars.keys()
        or "unused_supply" in results.data_vars.keys()
    ):
        results["unmet_demand"] = results.get("unmet_demand", 0) + results.get(
            "unused_supply", 0
        )

        results = results.drop_vars("unused_supply")

    return results



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

conversion_plus.py
~~~~~~~~~~~~~~~~~~

Conversion plus technology constraints.

"""

import pyomo.core as po  # pylint: disable=import-error

from calliope.backend.pyomo.util import (
    get_param,
    split_comma_list,
    get_conversion_plus_io,
)

ORDER = 20  # order in which to invoke constraints relative to other constraint files


def load_constraints(backend_model):
    sets = backend_model.__calliope_model_data["sets"]

    if "loc_techs_balance_conversion_plus_primary_constraint" in sets:
        backend_model.balance_conversion_plus_primary_constraint = po.Constraint(
            backend_model.loc_techs_balance_conversion_plus_primary_constraint,
            backend_model.timesteps,
            rule=balance_conversion_plus_primary_constraint_rule,
        )

    if "loc_techs_carrier_production_max_conversion_plus_constraint" in sets:
        backend_model.carrier_production_max_conversion_plus_constraint = po.Constraint(
            backend_model.loc_techs_carrier_production_max_conversion_plus_constraint,
            backend_model.timesteps,
            rule=carrier_production_max_conversion_plus_constraint_rule,
        )

    if "loc_techs_carrier_production_min_conversion_plus_constraint" in sets:
        backend_model.carrier_production_min_conversion_plus_constraint = po.Constraint(
            backend_model.loc_techs_carrier_production_min_conversion_plus_constraint,
            backend_model.timesteps,
            rule=carrier_production_min_conversion_plus_constraint_rule,
        )

    if "loc_techs_cost_var_conversion_plus_constraint" in sets:
        backend_model.cost_var_conversion_plus_constraint = po.Constraint(
            backend_model.costs,
            backend_model.loc_techs_cost_var_conversion_plus_constraint,
            backend_model.timesteps,
            rule=cost_var_conversion_plus_constraint_rule,
        )

    if "loc_techs_balance_conversion_plus_in_2_constraint" in sets:
        backend_model.balance_conversion_plus_in_2_constraint = po.Constraint(
            ["in_2"],
            backend_model.loc_techs_balance_conversion_plus_in_2_constraint,
            backend_model.timesteps,
            rule=balance_conversion_plus_tiers_constraint_rule,
        )

    if "loc_techs_balance_conversion_plus_in_3_constraint" in sets:
        backend_model.balance_conversion_plus_in_3_constraint = po.Constraint(
            ["in_3"],
            backend_model.loc_techs_balance_conversion_plus_in_3_constraint,
            backend_model.timesteps,
            rule=balance_conversion_plus_tiers_constraint_rule,
        )

    if "loc_techs_balance_conversion_plus_out_2_constraint" in sets:
        backend_model.balance_conversion_plus_out_2_constraint = po.Constraint(
            ["out_2"],
            backend_model.loc_techs_balance_conversion_plus_out_2_constraint,
            backend_model.timesteps,
            rule=balance_conversion_plus_tiers_constraint_rule,
        )

    if "loc_techs_balance_conversion_plus_out_3_constraint" in sets:
        backend_model.balance_conversion_plus_out_3_constraint = po.Constraint(
            ["out_3"],
            backend_model.loc_techs_balance_conversion_plus_out_3_constraint,
            backend_model.timesteps,
            rule=balance_conversion_plus_tiers_constraint_rule,
        )

    if "loc_tech_carrier_tiers_conversion_plus_zero_ratio_constraint" in sets:
        backend_model.conversion_plus_prod_con_to_zero_constraint = po.Constraint(
            backend_model.loc_tech_carrier_tiers_conversion_plus_zero_ratio_constraint,
            backend_model.timesteps,
            rule=conversion_plus_prod_con_to_zero_constraint_rule,
        )


def balance_conversion_plus_primary_constraint_rule(backend_model, loc_tech, timestep):
    """
    Balance energy carrier consumption and production for carrier_in and carrier_out

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{loc::tech::carrier \\in loc::tech::carriers_{out}}
            \\frac{\\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)}{
                carrier\\_ratio(loc::tech::carrier, `out')} =
            -1 * \\sum_{loc::tech::carrier \\in loc::tech::carriers_{in}} (
            \\boldsymbol{carrier_{con}}(loc::tech::carrier, timestep)
            * carrier\\_ratio(loc::tech::carrier, `in') * \\eta_{energy}(loc::tech, timestep))
            \\quad \\forall loc::tech \\in loc::techs_{conversion^{+}}, \\forall timestep \\in timesteps
    """
    model_data_dict = backend_model.__calliope_model_data["data"]

    loc_tech_carriers_out = split_comma_list(
        model_data_dict["lookup_loc_techs_conversion_plus"]["out", loc_tech]
    )
    loc_tech_carriers_in = split_comma_list(
        model_data_dict["lookup_loc_techs_conversion_plus"]["in", loc_tech]
    )

    energy_eff = get_param(backend_model, "energy_eff", (loc_tech, timestep))

    carrier_prod = []
    for loc_tech_carrier in loc_tech_carriers_out:
        carrier_ratio = get_param(
            backend_model, "carrier_ratios", ("out", loc_tech_carrier, timestep)
        )
        if po.value(carrier_ratio) != 0:
            carrier_prod.append(
                backend_model.carrier_prod[loc_tech_carrier, timestep] / carrier_ratio
            )

    carrier_con = sum(
        backend_model.carrier_con[loc_tech_carrier, timestep]
        * get_param(backend_model, "carrier_ratios", ("in", loc_tech_carrier, timestep))
        for loc_tech_carrier in loc_tech_carriers_in
    )

    return sum(carrier_prod) == -1 * carrier_con * energy_eff


def carrier_production_max_conversion_plus_constraint_rule(
    backend_model, loc_tech, timestep
):
    """
    Set maximum conversion_plus carrier production.

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{loc::tech::carrier \\in loc::tech::carriers_{out}}
            \\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)
            \\leq \\boldsymbol{energy_{cap}}(loc::tech) \\times timestep\\_resolution(timestep)
            \\quad \\forall loc::tech \\in loc::techs_{conversion^{+}},
            \\forall timestep \\in timesteps
    """
    model_data_dict = backend_model.__calliope_model_data["data"]

    timestep_resolution = backend_model.timestep_resolution[timestep]
    loc_tech_carriers_out = split_comma_list(
        model_data_dict["lookup_loc_techs_conversion_plus"]["out", loc_tech]
    )

    carrier_prod = sum(
        backend_model.carrier_prod[loc_tech_carrier, timestep]
        for loc_tech_carrier in loc_tech_carriers_out
    )

    return carrier_prod <= timestep_resolution * backend_model.energy_cap[loc_tech]


def carrier_production_min_conversion_plus_constraint_rule(
    backend_model, loc_tech, timestep
):
    """
    Set minimum conversion_plus carrier production.

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{loc::tech::carrier \\in loc::tech::carriers_{out}}
            \\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)
            \\leq \\boldsymbol{energy_{cap}}(loc::tech) \\times timestep\\_resolution(timestep)
            \\times energy_{cap, min use}(loc::tech)
            \\quad \\forall loc::tech \\in loc::techs_{conversion^{+}},
            \\forall timestep \\in timesteps
    """
    model_data_dict = backend_model.__calliope_model_data["data"]

    timestep_resolution = backend_model.timestep_resolution[timestep]
    min_use = get_param(backend_model, "energy_cap_min_use", (loc_tech, timestep))

    loc_tech_carriers_out = split_comma_list(
        model_data_dict["lookup_loc_techs_conversion_plus"]["out", loc_tech]
    )

    carrier_prod = sum(
        backend_model.carrier_prod[loc_tech_carrier, timestep]
        for loc_tech_carrier in loc_tech_carriers_out
    )

    return carrier_prod >= (
        timestep_resolution * backend_model.energy_cap[loc_tech] * min_use
    )


def cost_var_conversion_plus_constraint_rule(backend_model, cost, loc_tech, timestep):
    """
    Add time-varying conversion_plus technology costs

    .. container:: scrolling-wrapper

        .. math::

            \\boldsymbol{cost_{var}}(loc::tech, cost, timestep) =
            \\boldsymbol{carrier_{prod}}(loc::tech::carrier_{primary}, timestep)
            \\times timestep_{weight}(timestep) \\times cost_{om, prod}(loc::tech, cost, timestep)
            +
            \\boldsymbol{carrier_{con}}(loc::tech::carrier_{primary}, timestep)
            \\times timestep_{weight}(timestep) \\times cost_{om, con}(loc::tech, cost, timestep)
            \\quad \\forall loc::tech \\in loc::techs_{cost_{var}, conversion^{+}}
    """
    model_data_dict = backend_model.__calliope_model_data["data"]
    weight = backend_model.timestep_weights[timestep]

    loc_tech_carrier_con = model_data_dict["lookup_primary_loc_tech_carriers_in"][
        loc_tech
    ]

    loc_tech_carrier_prod = model_data_dict["lookup_primary_loc_tech_carriers_out"][
        loc_tech
    ]
    var_cost = 0

    if loc_tech_carrier_prod in backend_model.loc_tech_carriers_prod:
        cost_om_prod = get_param(
            backend_model, "cost_om_prod", (cost, loc_tech, timestep)
        )
        if cost_om_prod:
            var_cost += (
                cost_om_prod
                * weight
                * backend_model.carrier_prod[loc_tech_carrier_prod, timestep]
            )

    if loc_tech_carrier_con in backend_model.loc_tech_carriers_con:
        cost_om_con = get_param(
            backend_model, "cost_om_con", (cost, loc_tech, timestep)
        )
        if cost_om_con:
            var_cost += (
                cost_om_con
                * weight
                * -1
                * backend_model.carrier_con[loc_tech_carrier_con, timestep]
            )

    backend_model.cost_var_rhs[cost, loc_tech, timestep] = var_cost
    return (
        backend_model.cost_var[cost, loc_tech, timestep]
        == backend_model.cost_var_rhs[cost, loc_tech, timestep]
    )


def balance_conversion_plus_tiers_constraint_rule(
    backend_model, tier, loc_tech, timestep
):
    """
    Force all carrier_in_2/carrier_in_3 and carrier_out_2/carrier_out_3 to follow
    carrier_in and carrier_out (respectively).

    If `tier` in ['out_2', 'out_3']:

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{loc::tech::carrier \\in loc::tech::carriers_{out}} (
            \\frac{\\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)}{
                carrier\\_ratio(loc::tech::carrier, `out')} =
            \\sum_{loc::tech::carrier \\in loc::tech::carriers_{tier}} (
            \\frac{\\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)}{
                carrier\\_ratio(loc::tech::carrier, tier)}
            \\quad \\forall \\text { tier } \\in [`out_2', `out_3'], \\forall loc::tech
                \\in loc::techs_{conversion^{+}}, \\forall timestep \\in timesteps

    If `tier` in ['in_2', 'in_3']:

    .. container:: scrolling-wrapper

        .. math::

            \\sum_{loc::tech::carrier \\in loc::tech::carriers_{in}}
            \\frac{\\boldsymbol{carrier_{con}}(loc::tech::carrier, timestep)}{
                carrier\\_ratio(loc::tech::carrier, `in')} =
            \\sum_{loc::tech::carrier \\in loc::tech::carriers_{tier}}
            \\frac{\\boldsymbol{carrier_{prod}}(loc::tech::carrier, timestep)}{
                carrier\\_ratio(loc::tech::carrier, tier)}
            \\quad \\forall \\text{ tier } \\in [`in_2', `in_3'], \\forall loc::tech
                \\in loc::techs_{conversion^{+}}, \\forall timestep \\in timesteps
    """
    primary_tier, decision_variable = get_conversion_plus_io(backend_model, tier)
    model_data_dict = backend_model.__calliope_model_data["data"]

    loc_tech_carriers_1 = split_comma_list(
        model_data_dict["lookup_loc_techs_conversion_plus"][primary_tier, loc_tech]
    )
    loc_tech_carriers_2 = split_comma_list(
        model_data_dict["lookup_loc_techs_conversion_plus"][tier, loc_tech]
    )

    c_1 = []
    c_2 = []
    for loc_tech_carrier in loc_tech_carriers_1:
        carrier_ratio_1 = get_param(
            backend_model, "carrier_ratios", (primary_tier, loc_tech_carrier, timestep)
        )
        if po.value(carrier_ratio_1) != 0:
            c_1.append(decision_variable[loc_tech_carrier, timestep] / carrier_ratio_1)
    for loc_tech_carrier in loc_tech_carriers_2:
        carrier_ratio_2 = get_param(
            backend_model, "carrier_ratios", (tier, loc_tech_carrier, timestep)
        )
        if po.value(carrier_ratio_2) != 0:
            c_2.append(decision_variable[loc_tech_carrier, timestep] / carrier_ratio_2)
    if len(c_2) == 0:
        return po.Constraint.Skip
    else:
        return sum(c_1) == sum(c_2)


def conversion_plus_prod_con_to_zero_constraint_rule(
    backend_model, loc_tech_carrier_tier, timestep
):
    """
    Force any carrier production or consumption for a conversion plus technology to
    zero in timesteps where its carrier_ratio is zero
    """
    loc_tech_carrier, tier = loc_tech_carrier_tier.rsplit("::", 1)
    primary_tier, decision_variable = get_conversion_plus_io(backend_model, tier)

    carrier_ratio = get_param(
        backend_model, "carrier_ratios", (tier, loc_tech_carrier, timestep)
    )
    if po.value(carrier_ratio) == 0:
        return decision_variable[loc_tech_carrier, timestep] == 0
    else:
        return po.Constraint.Skip



"""
Copyright (C) since 2013 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

generate_runs.py
~~~~~~~~~~~~~~~~

Generate scripts to run multiple versions of the same model
in parallel on a cluster or sequentially on any machine.

"""

import os

import pandas as pd
from calliope.core import AttrDict


def generate_runs(model_file, scenarios=None, additional_args=None, override_dict=None):
    """
    Returns a list of "calliope run" invocations.

    ``scenarios`` must be specified as either a semicolon-separated
    list of scenarios or a semicolon-separated list of comma-separated
    individual override combinations, such as:

        ``override1,override2;override1,override3;...``

    If ``scenarios`` is not given, use all scenarios in the model
    configuration, and if no scenarios given in the model configuration,
    uses all individual overrides, one by one.

    """
    if scenarios is None:
        config = AttrDict.from_yaml(model_file)
        if override_dict:
            override = AttrDict.from_yaml_string(override_dict)
            config.union(override, allow_override=True, allow_replacement=True)

        if "scenarios" in config:
            runs = config.scenarios.keys()
        else:
            runs = config.overrides.keys()
    else:
        runs = scenarios.split(";")

    commands = []

    # len(str(x)) gives us the number of digits in x, for padding
    i_string = "{:0>" + str(len(str(len(runs)))) + "d}"

    for i, run in enumerate(runs):
        cmd = (
            (
                "calliope run {model} --scenario {scenario} "
                "--save_netcdf out_{i}_{scenario}.nc "
                "--save_plots plots_{i}_{scenario}.html"
            )
            .format(
                i=i_string.format(i + 1),
                model=model_file,
                scenario=run,
                override_dict=override_dict,
            )
            .strip()
        )

        if override_dict:
            cmd = cmd + ' --override_dict="{}"'.format(override_dict)

        if additional_args:
            cmd = cmd + " " + additional_args

        commands.append(cmd)

    return commands


def generate_bash_script(
    out_file, model_file, scenarios, additional_args=None, override_dict=None, **kwargs
):

    commands = generate_runs(model_file, scenarios, additional_args, override_dict)

    base_string = "    {i}) {cmd} ;;\n"
    lines_start = [
        "#!/bin/sh",
        "",
        "function process_case () {",
        '    case "$1" in',
        "",
    ]
    lines_end = [
        "    esac",
        "}",
        "",
        "if [[ $# -eq 0 ]] ; then",
        "    echo No parameter given, running all runs sequentially...",
        "    for i in $(seq 1 {}); do process_case $i; done".format(len(commands)),
        "else",
        "    echo Running run $1",
        "    process_case $1",
        "fi",
        "",
    ]

    lines_all = (
        lines_start
        + [base_string.format(i=i + 1, cmd=cmd) for i, cmd in enumerate(commands)]
        + lines_end
    )

    with open(out_file, "wb") as f:
        f.write(bytes("\n".join(lines_all), "UTF-8"))

    os.chmod(out_file, 0o755)

    return commands


def generate_bsub_script(
    out_file,
    model_file,
    scenarios,
    additional_args,
    override_dict,
    cluster_mem,
    cluster_time,
    cluster_threads=1,
    **kwargs,
):

    # We also need to generate the bash script to run on the cluster
    bash_out_file = out_file + ".array.sh"
    bash_out_file_basename = os.path.basename(bash_out_file)
    commands = generate_bash_script(
        bash_out_file, model_file, scenarios, additional_args, override_dict
    )

    lines = [
        "#!/bin/sh",
        "#BSUB -J calliope[1-{}]".format(len(commands)),
        "#BSUB -n {}".format(cluster_threads),
        '#BSUB -R "rusage[mem={}]"'.format(cluster_mem),
        "#BSUB -W {}".format(cluster_time),
        "#BSUB -r",  # Automatically restart failed jobs
        "#BSUB -o log_%I.log",
        "",
        "./" + bash_out_file_basename + " ${LSB_JOBINDEX}",
        "",
    ]

    with open(out_file, "wb") as f:
        f.write(bytes("\n".join(lines), "UTF-8"))


def generate_sbatch_script(
    out_file,
    model_file,
    scenarios,
    additional_args,
    override_dict,
    cluster_mem,
    cluster_time,
    cluster_threads=1,
    **kwargs,
):
    """
    SBATCH (SLURM) script generator.
    """

    # We also need to generate the bash script to run on the cluster
    bash_out_file = out_file + ".array.sh"
    bash_out_file_basename = os.path.basename(bash_out_file)
    commands = generate_bash_script(
        bash_out_file, model_file, scenarios, additional_args, override_dict
    )

    if ":" not in cluster_time:
        # Assuming time given as minutes, so needs changing to %H:%M%S
        cluster_time = pd.to_datetime(cluster_time, unit="m").strftime("%H:%M:%S")

    lines = [
        "#!/bin/bash",
        "#SBATCH -J calliope",  # Name of the job
        "#SBATCH --array=1-{}".format(len(commands)),  # How many jobs there are
        "#SBATCH --ntasks={}".format(cluster_threads),
        "#SBATCH --mem={}".format(cluster_mem),
        "#SBATCH --time={}".format(
            cluster_time
        ),  # How much wallclock time will be required
        "#SBATCH -o log_%a.log",
        "",
        "#! Optional add-ins for SBATCH (uncomment and add info as necessary):",
        "##SBATCH -A project_name",  # Which project should be charged'
        "##SBATCH --nodes=X",  # X whole nodes should be allocated'
        "##SBATCH -p partition_name",
        "",
        "#! Insert module load commands after this line, if needed:",
        "#! (Note: you can load this in ~.bashrc if you want them loaded every time you log in)",
        "",
        "#! module load gurobi (or glpk/cplex)",
        "#! module load miniconda3",
        "#! module load /path/to/miniconda3/envs/your_env_name/",
        "",
        "cd $SLURM_SUBMIT_DIR",
        "",
        "./" + bash_out_file_basename + " ${SLURM_ARRAY_TASK_ID}",
    ]

    with open(out_file, "wb") as f:
        f.write(bytes("\n".join(lines), "UTF-8"))


def generate_windows_script(
    out_file, model_file, scenarios, additional_args=None, override_dict=None, **kwargs
):

    commands = generate_runs(model_file, scenarios, additional_args, override_dict)

    # \r\n are Windows line endings
    base_string = 'echo "Run {i}"\r\n{cmd}\r\n'
    lines_start = [
        "@echo off",
        "",
    ]

    lines_all = lines_start + [
        base_string.format(i=i + 1, cmd=cmd) for i, cmd in enumerate(commands)
    ]

    with open(out_file, "wb") as f:
        f.write(bytes("\r\n".join(lines_all), "UTF-8"))

    os.chmod(out_file, 0o755)

    return commands


_KINDS = {
    "bash": generate_bash_script,
    "bsub": generate_bsub_script,
    "windows": generate_windows_script,
    "sbatch": generate_sbatch_script,
}


def generate(kind, **kwargs):
    _KINDS[kind](**kwargs)



from io import StringIO
import os

import pytest
import numpy as np
import tempfile
import ruamel.yaml as ruamel_yaml

from calliope.core.attrdict import AttrDict, _MISSING
from calliope.test.common.util import check_error_or_warning


class TestAttrDict:
    @pytest.fixture
    def regular_dict(self):
        d = {
            "a": 1,
            "b": 2,
            "c": {"x": "foo", "y": "bar", "z": {"I": 1, "II": 2}},
            "d": None,
        }
        return d

    setup_string = """
        # a comment
        a: 1
        b: 2
        # a comment about `c`
        c:  # a comment inline with `c`
            x: foo  # a comment on foo

            #
            y: bar  #
            z:
                I: 1
                II: 2
        d:
    """

    @pytest.fixture
    def yaml_file(self):
        return StringIO(self.setup_string)

    @pytest.fixture
    def yaml_string(self):
        return self.setup_string

    @pytest.fixture
    def attr_dict(self, regular_dict):
        d = regular_dict
        return AttrDict(d)

    def test_missing_nonzero(self):
        assert _MISSING is not True
        assert _MISSING is not False
        assert _MISSING is not None
        assert _MISSING.__nonzero__() is False

    def test_init_from_nondict(self):
        with pytest.raises(ValueError) as excinfo:
            d = AttrDict("foo")
        assert check_error_or_warning(excinfo, "Must pass a dict to AttrDict")

    def test_init_from_dict(self, regular_dict):
        d = AttrDict(regular_dict)
        assert d.a == 1

    def test_init_from_dict_with_nested_keys(self):
        d = AttrDict({"foo.bar.baz": 1})
        assert d.foo.bar.baz == 1

    def test_from_yaml_fobj(self, yaml_file):
        d = AttrDict.from_yaml(yaml_file)
        assert d.a == 1
        assert d.c.z.II == 2

    def test_from_yaml_path(self):
        this_path = os.path.dirname(__file__)
        yaml_path = os.path.join(this_path, "common", "yaml_file.yaml")
        d = AttrDict.from_yaml(yaml_path)
        assert d.a == 1
        assert d.c.z.II == 2

    def test_from_yaml_string(self, yaml_string):
        d = AttrDict.from_yaml_string(yaml_string)
        assert d.a == 1
        assert d.c.z.II == 2

    def test_from_yaml_string_dot_strings(self):
        yaml_string = "a.b.c: 1\na.b.foo: 2"
        d = AttrDict.from_yaml_string(yaml_string)
        assert d.a.b.c == 1
        assert d.a.b.foo == 2

    def test_from_yaml_string_dot_strings_duplicate(self):
        yaml_string = "a.b.c: 1\na.b.c: 2"
        with pytest.raises(ruamel_yaml.constructor.DuplicateKeyError):
            AttrDict.from_yaml_string(yaml_string)

    def test_simple_invalid_yaml(self):
        yaml_string = "1 this is not valid yaml"
        with pytest.raises(ValueError) as excinfo:
            AttrDict.from_yaml_string(yaml_string)
        assert check_error_or_warning(excinfo, "Could not parse <yaml string> as YAML")

    def test_parser_error(self):
        with pytest.raises(ruamel_yaml.YAMLError):
            AttrDict.from_yaml_string(
                """
            foo: bar
            baz: 1
                - foobar
                bar: baz

            """
            )

    def test_order_of_subdicts(self):
        d = AttrDict.from_yaml_string(
            """
            A.B.C: 10
            A.B:
                E: 20
        """
        )
        assert d.A.B.C == 10
        assert d.A.B.E == 20

    def test_dot_access_first(self, attr_dict):
        d = attr_dict
        assert d.a == 1

    def test_dot_access_second(self, attr_dict):
        d = attr_dict
        assert d.c.x == "foo"

    def test_dot_access_list(self):
        d = AttrDict.from_yaml_string("a: [{x: 1}, {y: 2}]")
        assert d.a[0].x == 1

    def test_set_key_first(self, attr_dict):
        d = attr_dict
        d.set_key("a", 2)
        assert d.a == 2

    def test_set_key_second(self, attr_dict):
        d = attr_dict
        d.set_key("c.x", "baz")
        assert d.c.x == "baz"

    def test_set_key_multiple_inexisting(self, attr_dict):
        d = attr_dict
        d.set_key("c.l.o.h.a", "foo")
        assert d.c.l.o.h.a == "foo"

    def test_set_key_nested_on_string(self, attr_dict):
        d = attr_dict
        with pytest.raises(KeyError):
            d.set_key("a.foo", "bar")

    def test_set_key_nested_on_none(self, attr_dict):
        d = attr_dict
        assert d["d"] is None
        d.set_key("d.foo", "bar")
        assert d.d.foo == "bar"

    def test_pass_regular_dict_to_set_key(self, attr_dict):
        # Regular dicts get turned into AttrDicts when using
        # assignment through set_key()
        attr_dict.set_key("c.z.newkey", {"foo": 1, "doo": 2})
        assert isinstance(attr_dict.get_key("c.z.newkey"), AttrDict)
        assert attr_dict.get_key("c.z.newkey.foo") == 1

    def test_get_subkey_from_nested_non_attrdict(self, attr_dict):
        # Directly assigning a dict means it is not modified
        # but it breaks get_key with nested keys
        attr_dict["c"]["z"]["newkey"] = {"foo": 1, "doo": 2}

        with pytest.raises(AttributeError) as excinfo:
            attr_dict.get_key("c.z.newkey.foo")

        assert check_error_or_warning(
            excinfo, "'dict' object has no attribute 'get_key'"
        )

    def test_get_key_first(self, attr_dict):
        d = attr_dict
        assert d.get_key("a") == 1

    def test_get_key_second(self, attr_dict):
        d = attr_dict
        assert d.get_key("c.x") == "foo"

    def test_get_key_inexistant(self, attr_dict):
        d = attr_dict
        with pytest.raises(KeyError):
            d.get_key("foo")

    def test_get_key_second_inexistant(self, attr_dict):
        d = attr_dict
        with pytest.raises(KeyError):
            d.get_key("foo.bar")

    def test_get_key_default(self, attr_dict):
        d = attr_dict
        assert d.get_key("c.x", default="bar") == "foo"

    def test_get_key_inexistant_default(self, attr_dict):
        d = attr_dict
        assert d.get_key("foo", default="baz") == "baz"

    def test_get_key_second_inexistant_default(self, attr_dict):
        d = attr_dict
        assert d.get_key("foo.bar", default="baz") == "baz"

    def test_get_key_second_nondict_default(self, attr_dict):
        d = attr_dict
        assert d.get_key("c.x.foo", default="baz") == "baz"

    def test_get_key_inexistant_default_false(self, attr_dict):
        d = attr_dict
        assert d.get_key("foo", default=False) is False

    def test_get_key_second_inexistant_default_false(self, attr_dict):
        d = attr_dict
        assert d.get_key("foo.bar", default=False) is False

    def test_as_dict(self, attr_dict):
        d = attr_dict
        dd = d.as_dict()
        assert dd["a"] == 1
        assert dd["c"]["x"] == "foo"

    def test_as_dict_with_sublists(self):
        d = AttrDict.from_yaml_string("a: [{x: 1}, {y: 2}]")
        dd = d.as_dict()
        assert dd["a"][0]["x"] == 1
        assert isinstance(dd["a"][0], dict)  # Not AttrDict!

    def test_as_dict_flat(self, attr_dict):
        dd = attr_dict.as_dict(flat=True)
        assert dd["c.x"] == "foo"

    def test_keys_nested_as_list(self, attr_dict):
        d = attr_dict
        dd = d.keys_nested()
        assert dd == ["a", "b", "c.x", "c.y", "c.z.I", "c.z.II", "d"]

    def test_keys_nested_as_dict(self, attr_dict):
        d = attr_dict
        dd = d.keys_nested(subkeys_as="dict")
        assert dd == ["a", "b", {"c": ["x", "y", {"z": ["I", "II"]}]}, "d"]

    def test_union(self, attr_dict):
        d = attr_dict
        d_new = AttrDict()
        d_new.set_key("c.z.III", "foo")
        d.union(d_new)
        assert d.c.z.III == "foo"
        assert d.c.z.I == 1

    def test_union_duplicate_keys(self, attr_dict):
        d = attr_dict
        d_new = AttrDict()
        d_new.set_key("c.z.II", "foo")
        with pytest.raises(KeyError):
            d.union(d_new)

    def test_union_replacement(self, attr_dict):
        d = attr_dict
        d_new = AttrDict.from_yaml_string(
            """
            c: {_REPLACE_: foo}
        """
        )
        d.union(d_new, allow_override=True, allow_replacement=True)
        assert d.c == "foo"

    def test_union_empty_dicts(self, attr_dict):
        d = attr_dict
        d_new = AttrDict({"1": {"foo": {}}, "baz": {"bar": {}}})
        d.union(d_new)
        assert len(d.baz.bar.keys()) == 0

    def test_del_key_single(self, attr_dict):
        attr_dict.del_key("c")
        assert "c" not in attr_dict

    def test_del_key_nested(self, attr_dict):
        attr_dict.del_key("c.z.I")
        assert "I" not in attr_dict.c.z

    def test_to_yaml(self, yaml_file):
        d = AttrDict.from_yaml(yaml_file)
        d.set_key("numpy.some_int", np.int32(10))
        d.set_key("numpy.some_float", np.float64(0.5))
        d.a_list = [0, 1, 2]
        with tempfile.TemporaryDirectory() as tempdir:
            out_file = os.path.join(tempdir, "test.yaml")
            d.to_yaml(out_file)

            with open(out_file, "r") as f:
                result = f.read()

            assert "some_int: 10" in result
            assert "some_float: 0.5" in result
            assert "a_list:\n- 0\n- 1\n- 2" in result

    def test_to_yaml_string(self, yaml_file):
        d = AttrDict.from_yaml(yaml_file)
        result = d.to_yaml()
        assert "a: 1" in result

    def test_import_must_be_list(self):
        yaml_string = """
            import: 'somefile.yaml'
        """
        with pytest.raises(ValueError) as excinfo:
            AttrDict.from_yaml_string(yaml_string, resolve_imports=True)
        assert check_error_or_warning(excinfo, "`import` must be a list.")

    def test_do_not_resolve_imports(self):
        yaml_string = """
            import: ['somefile.yaml']
        """
        d = AttrDict.from_yaml_string(yaml_string, resolve_imports=False)
        # Should not raise an error about a missing file, as we ask for
        # imports not to be resolved
        assert d["import"] == ["somefile.yaml"]

    def test_nested_import(self, yaml_file):
        with tempfile.TemporaryDirectory() as tempdir:
            imported_file = os.path.join(tempdir, "test_import.yaml")
            imported_yaml = """
                somekey: 1
                anotherkey: 2
            """
            with open(imported_file, "w") as f:
                f.write(imported_yaml)

            yaml_string = """
                foobar:
                    import:
                        - {}
                foo:
                    bar: 1
                    baz: 2
                    3:
                        4: 5
            """.format(
                imported_file
            )

            d = AttrDict.from_yaml_string(yaml_string, resolve_imports="foobar")

        assert "foobar.somekey" in d.keys_nested()
        assert d.get_key("foobar.anotherkey") == 2
